"""
Datasets Repository
"""

import copy
import json
import logging
import os
import sys
import tempfile
import time
import zipfile
from pathlib import Path
from typing import Generator, Optional, Union

import tqdm

from .. import _api_reference, entities, exceptions, miscellaneous, PlatformException, repositories, services
from ..entities.dataset import ExportType, OutputExportType
from ..services import service_defaults
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')

MAX_ITEMS_PER_SUBSET = 50000
DOWNLOAD_ANNOTATIONS_MAX_ITEMS_PER_SUBSET = 1000

class Datasets:
    """
    Datasets Repository

    The Datasets class allows the user to manage datasets. Read more about datasets in our `documentation <https://dataloop.ai/docs/dataset>`_ and `SDK documentation <https://developers.dataloop.ai/tutorials/data_management/manage_datasets/chapter/>`_.
    """

    def __init__(self, client_api: ApiClient, project: entities.Project = None):
        self._client_api = client_api
        self._project = project

    ############
    # entities #
    ############
    @property
    def project(self) -> entities.Project:
        if self._project is None:
            # try get checkout
            project = self._client_api.state_io.get('project')
            if project is not None:
                self._project = entities.Project.from_json(_json=project, client_api=self._client_api)
        if self._project is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Cannot perform action WITHOUT Project entity in Datasets repository.'
                        ' Please checkout or set a project')
        assert isinstance(self._project, entities.Project)
        return self._project

    @project.setter
    def project(self, project: entities.Project):
        if not isinstance(project, entities.Project):
            raise ValueError('Must input a valid Project entity')
        self._project = project

    ###########
    # methods #
    ###########
    def __get_from_cache(self) -> entities.Dataset:
        dataset = self._client_api.state_io.get('dataset')
        if dataset is not None:
            dataset = entities.Dataset.from_json(_json=dataset,
                                                 client_api=self._client_api,
                                                 datasets=self,
                                                 project=self._project)
        return dataset

    def __get_by_id(self, dataset_id) -> entities.Dataset:
        success, response = self._client_api.gen_request(req_type='get',
                                                         path='/datasets/{}'.format(dataset_id))
        if dataset_id is None or dataset_id == '':
            raise exceptions.PlatformException('400', 'Please checkout a dataset')

        if success:
            dataset = entities.Dataset.from_json(client_api=self._client_api,
                                                 _json=response.json(),
                                                 datasets=self,
                                                 project=self._project)
        else:
            raise exceptions.PlatformException(response)
        return dataset

    def __get_by_identifier(self, identifier=None) -> entities.Dataset:
        datasets = self.list()
        datasets_by_name = [dataset for dataset in datasets if identifier in dataset.name or identifier in dataset.id]
        if len(datasets_by_name) == 1:
            return datasets_by_name[0]
        elif len(datasets_by_name) > 1:
            raise Exception('Multiple datasets with this name exist')
        else:
            raise Exception("Dataset not found")

    def _bulid_folder_filter(self, folder_path, filters=None):
        if filters is None:
            filters = entities.Filters()
            filters._user_query = 'false'
        if not folder_path.startswith('/'):
            folder_path = '/' + folder_path
        filters.add(field='dir', values=folder_path, method=entities.FiltersMethod.OR)
        if not folder_path.endswith('*'):
            if not folder_path.endswith('/'):
                folder_path += '/'
            filters.add(field='dir', values=folder_path + '*', method=entities.FiltersMethod.OR)
        return filters

    def _get_binaries_dataset(self):
        filters = entities.Filters(resource=entities.FiltersResource.DATASET)
        filters.add(field='name', values='Binaries')
        filters.system_space = True
        datasets = self.list(filters=filters)
        if len(datasets) == 0:
            # empty list
            raise exceptions.PlatformException('404', 'Dataset not found. Name: "Binaries"')
            # dataset = None
        elif len(datasets) > 1:
            raise exceptions.PlatformException('400', 'More than one dataset with same name.')
        else:
            dataset = datasets[0]
        return dataset

    def _resolve_dataset_id(self, dataset, dataset_name, dataset_id):
        if dataset is None and dataset_name is None and dataset_id is None:
            raise ValueError('Must provide dataset, dataset name or dataset id')
        if dataset_id is None:
            if dataset is None:
                dataset = self.get(dataset_name=dataset_name)
            dataset_id = dataset.id
        return dataset_id

    @staticmethod
    def _save_item_json_file(item_data, base_path: Path, export_version=None):
        """
        Save a single item's JSON data to a file, creating the directory structure as needed.
        
        :param dict item_data: The item data dictionary (must have 'filename' key)
        :param Path base_path: Base directory path where JSON files should be saved
        :param entities.ExportVersion export_version: Optional export version (V1 or V2) affecting filename handling
        :return: Path to the saved JSON file
        :rtype: Path
        """
        # Get filename and remove leading slash
        filename = item_data.get('filename', '')
        if not filename:
            raise ValueError("item_data must have a 'filename' key")
        filename = filename.lstrip('/')
        
        # Determine relative JSON path based on export version
        if export_version == entities.ExportVersion.V1:
            # V1: Replace extension with .json (e.g., "file.jpg" -> "file.json")
            rel_json_path = str(Path(filename).with_suffix('.json'))
        elif export_version == entities.ExportVersion.V2:
            # V2: Append .json (e.g., "file.jpg" -> "file.jpg.json")
            rel_json_path = filename + '.json'
        else:
            # Default/None: Replace extension with .json (backward compatible with section 1)
            rel_json_path = os.path.splitext(filename)[0] + '.json'
        
        # Remove leading slash if present
        if rel_json_path.startswith('/'):
            rel_json_path = rel_json_path[1:]
        
        # Build output path
        out_path = base_path / rel_json_path
        
        # Create parent directories
        out_path.parent.mkdir(parents=True, exist_ok=True)
        
        # Write JSON file
        try:
            with open(out_path, 'w') as outf:
                json.dump(item_data, outf, indent=2)
        except Exception:
            logger.exception(f'Failed writing export item JSON to {out_path}')
            raise
        
        return out_path

    @staticmethod
    def _build_payload(filters, include_feature_vectors, include_annotations,
                       export_type, annotation_filters, feature_vector_filters, dataset_lock, lock_timeout_sec, export_summary):
        valid_list = [e.value for e in entities.ExportType]
        valid_types = ', '.join(valid_list)
        if export_type not in ['json', 'zip']:
            raise ValueError('export_type must be one of the following: {}'.format(valid_types))
        payload = {'exportType': export_type}
        if filters is None:
            filters = entities.Filters()

        if isinstance(filters, entities.Filters):
            payload['itemsQuery'] = {'filter': filters.prepare()['filter'], 'join': filters.prepare().get("join", {})}
        elif isinstance(filters, dict):
            payload['itemsQuery'] = filters
        else:
            raise exceptions.BadRequest(message='filters must be of type dict or Filters', status_code=500)

        payload['itemsVectorQuery'] = {}
        if include_feature_vectors:
            payload['includeItemVectors'] = True
            payload['itemsVectorQuery']['select'] = {"datasetId": 1, 'featureSetId': 1, 'value': 1}

        if feature_vector_filters is not None:
            payload['itemsVectorQuery']['filter'] = feature_vector_filters.prepare()['filter']

        payload['annotations'] = {"include": include_annotations, "convertSemantic": False}

        if annotation_filters is not None:
            payload['annotationsQuery'] = annotation_filters.prepare()

        if dataset_lock:
            payload['datasetLock'] = dataset_lock

        if export_summary:
            payload['summary'] = export_summary

        if lock_timeout_sec:
            payload['lockTimeoutSec'] = lock_timeout_sec

        return payload

    def _download_exported_item(self, item_id, export_type, local_path=None, unzip=True):
        logger.debug(f"start downloading exported item {item_id} with export_type {export_type} and local_path {local_path} and unzip {unzip}")
        export_item = repositories.Items(client_api=self._client_api).get(item_id=item_id)
        export_item_path = export_item.download(local_path=local_path)        

        # Common validation check for both JSON and other export types
        if isinstance(export_item_path, list) or not os.path.isfile(export_item_path):
            raise exceptions.PlatformException(
                error='404',
                message='error downloading annotation zip file. see above for more information. item id: {!r}'.format(
                    export_item.id))

        result = None
        if unzip is False or export_type == entities.ExportType.JSON:
            result = export_item_path
        else:
            try:
                miscellaneous.Zipping.unzip_directory(zip_filename=export_item_path,
                                                        to_directory=local_path)
                result = local_path
            except Exception as e:
                logger.warning("Failed to extract zip file error: {}".format(e))
            finally:
                # cleanup only for zip files to avoid removing needed results
                if isinstance(export_item_path, str) and os.path.isfile(export_item_path):
                    os.remove(export_item_path)
        logger.debug(f"end downloading, result {result}")
        return result

    @property
    def platform_url(self):
        return self._client_api._get_resource_url("projects/{}/datasets".format(self.project.id))

    def open_in_web(self,
                    dataset_name: str = None,
                    dataset_id: str = None,
                    dataset: entities.Dataset = None):
        """
        Open the dataset in web platform.

        **Prerequisites**: You must be an *owner* or *developer* to use this method.

        :param str dataset_name: The Name of the dataset
        :param str dataset_id: The Id of the dataset
        :param dtlpy.entities.dataset.Dataset dataset: dataset object

        **Example**:

        .. code-block:: python

            project.datasets.open_in_web(dataset_id='dataset_id')
        """
        if dataset_name is not None:
            dataset = self.get(dataset_name=dataset_name)
        if dataset is not None:
            dataset.open_in_web()
        elif dataset_id is not None:
            self._client_api._open_in_web(url=f'{self.platform_url}/{dataset_id}/items')
        else:
            self._client_api._open_in_web(url=self.platform_url)

    def checkout(self,
                 identifier: str = None,
                 dataset_name: str = None,
                 dataset_id: str = None,
                 dataset: entities.Dataset = None):
        """
        Checkout (switch) to a dataset to work on it.

        **Prerequisites**: You must be an *owner* or *developer* to use this method.

        You must provide at least ONE of the following params: dataset_id, dataset_name.

        :param str identifier: project name or partial id that you wish to switch
        :param str dataset_name: The Name of the dataset
        :param str dataset_id: The Id of the dataset
        :param dtlpy.entities.dataset.Dataset dataset: dataset object

        **Example**:

        .. code-block:: python

            project.datasets.checkout(dataset_id='dataset_id')
        """
        if dataset is None:
            if dataset_id is not None or dataset_name is not None:
                try:
                    dataset = self.project.datasets.get(dataset_name=dataset_name, dataset_id=dataset_id)
                except exceptions.MissingEntity:
                    dataset = self.get(dataset_id=dataset_id, dataset_name=dataset_name)
            elif identifier is not None:
                dataset = self.__get_by_identifier(identifier=identifier)
            else:
                raise exceptions.PlatformException(error='400',
                                                   message='Must provide partial/full id/name to checkout')
        self._client_api.state_io.put('dataset', dataset.to_json())
        logger.info('Checked out to dataset {}'.format(dataset.name))

    @_api_reference.add(path='/datasets/query', method='post')
    def list(self, name=None, creator=None, filters: entities.Filters = None) -> miscellaneous.List[entities.Dataset]:
        """
        List all datasets.

        **Prerequisites**: You must be an *owner* or *developer* to use this method.

        :param str name: list by name
        :param str creator: list by
        :param dtlpy.entities.filters.Filters filters: Filters entity containing filters parameters
        :return: List of datasets
        :rtype: list

        **Example**:

        .. code-block:: python
            filters = dl.Filters(resource='datasets')
            filters.add(field='readonly', values=False)
            datasets = project.datasets.list(filters=filters)
        """
        if filters is None:
            filters = entities.Filters(resource=entities.FiltersResource.DATASET)
        # assert type filters
        elif not isinstance(filters, entities.Filters):
            raise exceptions.PlatformException(error='400',
                                               message='Unknown filters type: {!r}'.format(type(filters)))
        if filters.resource != entities.FiltersResource.DATASET:
            raise exceptions.PlatformException(
                error='400',
                message='Filters resource must to be FiltersResource.DATASET. Got: {!r}'.format(filters.resource))

        url = '/datasets/query'

        if name is not None:
            filters.add(field='name', values=name)
        if creator is not None:
            filters.add(field='creator', values=creator)
        if self._project is not None:
            filters.context = {"projects": [self._project.id]}
        filters.page_size = 1000
        filters.page = 0
        datasets = list()
        while True:
            success, response = self._client_api.gen_request(req_type='POST',
                                                             json_req=filters.prepare(),
                                                             path=url,
                                                             headers={'user_query': filters._user_query})
            if success:
                pool = self._client_api.thread_pools('entity.create')
                datasets_json = response.json()['items']
                jobs = [None for _ in range(len(datasets_json))]
                # return triggers list
                for i_dataset, dataset in enumerate(datasets_json):
                    jobs[i_dataset] = pool.submit(entities.Dataset._protected_from_json,
                                                  **{'client_api': self._client_api,
                                                     '_json': dataset,
                                                     'datasets': self,
                                                     'project': self.project})

                # get all results
                results = [j.result() for j in jobs]
                # log errors
                _ = [logger.warning(r[1]) for r in results if r[0] is False]
                # return good jobs
                datasets.extend([r[1] for r in results if r[0] is True])
                if response.json()['hasNextPage'] is True:
                    filters.page += 1
                else:
                    break
            else:
                raise exceptions.PlatformException(response)
        datasets = miscellaneous.List(datasets)
        return datasets

    @_api_reference.add(path='/datasets/{id}', method='get')
    def get(self,
            dataset_name: str = None,
            dataset_id: str = None,
            checkout: bool = False,
            fetch: bool = None
            ) -> entities.Dataset:
        """
        Get dataset by name or id.

        **Prerequisites**: You must be an *owner* or *developer* to use this method.

        You must provide at least ONE of the following params: dataset_id, dataset_name.

        :param str dataset_name: optional - search by name
        :param str dataset_id: optional - search by id
        :param bool checkout: set the dataset as a default dataset object (cookies)
        :param bool fetch: optional - fetch entity from platform (True), default taken from cookie
        :return: Dataset object
        :rtype: dtlpy.entities.dataset.Dataset

        **Example**:

        .. code-block:: python

            dataset = project.datasets.get(dataset_id='dataset_id')
        """
        if fetch is None:
            fetch = self._client_api.fetch_entities

        if dataset_id is None and dataset_name is None:
            dataset = self.__get_from_cache()
            if dataset is None:
                raise exceptions.PlatformException(
                    error='400',
                    message='No checked-out Dataset was found, must checkout or provide an identifier in inputs')
        elif fetch:
            if dataset_id is not None and dataset_id != '':
                dataset = self.__get_by_id(dataset_id)
                # verify input dataset name is same as the given id
                if dataset_name is not None and dataset.name != dataset_name:
                    logger.warning(
                        "Mismatch found in datasets.get: dataset_name is different then dataset.name: "
                        "{!r} != {!r}".format(
                            dataset_name,
                            dataset.name))
            elif dataset_name is not None:
                datasets = self.list(name=dataset_name)
                if not datasets:
                    # empty list
                    raise exceptions.PlatformException('404', 'Dataset not found. Name: {!r}'.format(dataset_name))
                    # dataset = None
                elif len(datasets) > 1:
                    raise exceptions.PlatformException('400', 'More than one dataset with same name.')
                else:
                    dataset = datasets[0]
            else:
                raise exceptions.PlatformException(
                    error='404',
                    message='No input and no checked-out found')
        else:
            dataset = entities.Dataset.from_json(_json={'id': dataset_id,
                                                        'name': dataset_id},
                                                 client_api=self._client_api,
                                                 datasets=self,
                                                 project=self._project,
                                                 is_fetched=False)
        assert isinstance(dataset, entities.Dataset)
        if checkout:
            self.checkout(dataset=dataset)
        return dataset

    @_api_reference.add(path='/datasets/{id}', method='delete')
    def delete(self,
               dataset_name: str = None,
               dataset_id: str = None,
               sure: bool = False,
               really: bool = False):
        """
        Delete a dataset forever!

        **Prerequisites**: You must be an *owner* or *developer* to use this method.

        **Example**:

        .. code-block:: python

            is_deleted = project.datasets.delete(dataset_id='dataset_id', sure=True, really=True)

        :param str dataset_name: optional - search by name
        :param str dataset_id: optional - search by id
        :param bool sure: Are you sure you want to delete?
        :param bool really: Really really sure?
        :return: True is success
        :rtype: bool
        """
        if sure and really:
            dataset = self.get(dataset_name=dataset_name, dataset_id=dataset_id)
            success, response = self._client_api.gen_request(req_type='delete',
                                                             path='/datasets/{}'.format(dataset.id))
            if not success:
                raise exceptions.PlatformException(response)
            logger.info('Dataset {!r} was deleted successfully'.format(dataset.name))
            return True
        else:
            raise exceptions.PlatformException(
                error='403',
                message='Cant delete dataset from SDK. Please login to platform to delete')

    @_api_reference.add(path='/datasets/{id}', method='patch')
    def update(self,
               dataset: entities.Dataset,
               system_metadata: bool = False,
               patch: dict = None
               ) -> entities.Dataset:
        """
        Update dataset field.

        **Prerequisites**: You must be an *owner* or *developer* to use this method.

        :param dtlpy.entities.dataset.Dataset dataset: dataset object
        :param bool system_metadata: True, if you want to change metadata system
        :param dict patch: Specific patch request
        :return: Dataset object
        :rtype: dtlpy.entities.dataset.Dataset

        **Example**:

        .. code-block:: python

            dataset = project.datasets.update(dataset='dataset_entity')
        """
        url_path = '/datasets/{}'.format(dataset.id)
        if system_metadata:
            url_path += '?system=true'

        if patch is None:
            patch = dataset.to_json()

        success, response = self._client_api.gen_request(req_type='patch',
                                                         path=url_path,
                                                         json_req=patch)
        if success:
            logger.info('Dataset was updated successfully')
            return dataset
        else:
            raise exceptions.PlatformException(response)

    @_api_reference.add(path='/datasets/{id}/unlock', method='patch')
    def unlock(self, dataset: entities.Dataset ) -> entities.Dataset:
        """
        Unlock dataset.

        **Prerequisites**: You must be an *owner* or *developer* to use this method.

        :param dtlpy.entities.dataset.Dataset dataset: dataset object
        :return: Dataset object
        :rtype: dtlpy.entities.dataset.Dataset

        **Example**:

        .. code-block:: python

            dataset = project.datasets.unlock(dataset='dataset_entity')
        """
        url_path = '/datasets/{}/unlock'.format(dataset.id)

        success, response = self._client_api.gen_request(req_type='patch', path=url_path)
        if success:
            logger.info('Dataset was unlocked successfully')
            return dataset
        else:
            raise exceptions.PlatformException(response)

    @_api_reference.add(path='/datasets/{id}/directoryTree', method='get')
    def directory_tree(self,
                       dataset: entities.Dataset = None,
                       dataset_name: str = None,
                       dataset_id: str = None):
        """
        Get dataset's directory tree.

        **Prerequisites**: You must be an *owner* or *developer* to use this method.

        You must provide at least ONE of the following params: dataset, dataset_name, dataset_id.

        :param dtlpy.entities.dataset.Dataset dataset: dataset object
        :param str dataset_name: The Name of the dataset
        :param str dataset_id: The Id of the dataset
        :return: DirectoryTree

        **Example**:

        .. code-block:: python
            directory_tree = dataset.directory_tree
            directory_tree = project.datasets.directory_tree(dataset='dataset_entity')
        """
        dataset_id = self._resolve_dataset_id(dataset, dataset_name, dataset_id)

        url_path = '/datasets/{}/directoryTree'.format(dataset_id)

        success, response = self._client_api.gen_request(req_type='get',
                                                         path=url_path)

        if success:
            return entities.DirectoryTree(_json=response.json())
        else:
            raise exceptions.PlatformException(response)

    @_api_reference.add(path='/datasets/{id}/clone', method='post')
    def clone(self,
              dataset_id: str,
              clone_name: str = None,
              filters: entities.Filters = None,
              with_items_annotations: bool = True,
              with_metadata: bool = True,
              with_task_annotations_status: bool = True,
              dst_dataset_id: str = None,
              target_directory: str = None):
        """
        Clone a dataset. Read more about cloning datatsets and items in our `documentation <https://dataloop.ai/docs/clone-merge-dataset#cloned-dataset>`_ and `SDK documentation <https://developers.dataloop.ai/tutorials/data_management/data_versioning/chapter/>`_.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param str dataset_id: id of the dataset you wish to clone
        :param str clone_name: new dataset name
        :param dtlpy.entities.filters.Filters filters: Filters entity or a query dict
        :param bool with_items_annotations: true to clone with items annotations
        :param bool with_metadata: true to clone with metadata
        :param bool with_task_annotations_status: true to clone with task annotations' status
        :param str dst_dataset_id: destination dataset id
        :param str target_directory: target directory
        :return: dataset object
        :rtype: dtlpy.entities.dataset.Dataset

        **Example**:

        .. code-block:: python

            dataset = project.datasets.clone(dataset_id='dataset_id',
                                  clone_name='dataset_clone_name',
                                  with_metadata=True,
                                  with_items_annotations=False,
                                  with_task_annotations_status=False)
        """
        if clone_name is None and dst_dataset_id is None:
            raise exceptions.PlatformException('400', 'Must provide clone name or destination dataset id')
        if filters is None:
            filters = entities.Filters()
            filters._user_query = 'false'
        elif not isinstance(filters, entities.Filters):
            raise exceptions.PlatformException(
                error='400',
                message='"filters" must be a dl.Filters entity. got: {!r}'.format(type(filters)))

        copy_filters = copy.deepcopy(filters)
        if copy_filters.has_field('hidden'):
            copy_filters.pop('hidden')

        if target_directory is not None and not target_directory.startswith('/'):
            target_directory = '/' + target_directory

        payload = {
            "name": clone_name,
            "filter": copy_filters.prepare(),
            "cloneDatasetParams": {
                "withItemsAnnotations": with_items_annotations,
                "withMetadata": with_metadata,
                "withTaskAnnotationsStatus": with_task_annotations_status,
                "targetDirectory": target_directory
            }
        }
        if dst_dataset_id is not None:
            payload['cloneDatasetParams']['targetDatasetId'] = dst_dataset_id
        success, response = self._client_api.gen_request(req_type='post',
                                                         path='/datasets/{}/clone'.format(dataset_id),
                                                         json_req=payload,
                                                         headers={'user_query': filters._user_query})

        if not success:
            raise exceptions.PlatformException(response)

        command = entities.Command.from_json(_json=response.json(),
                                             client_api=self._client_api)
        command = command.wait()

        if 'returnedModelId' not in command.spec:
            raise exceptions.PlatformException(error='400',
                                               message="returnedModelId key is missing in command response: {!r}"
                                               .format(response))
        return self.get(dataset_id=command.spec['returnedModelId'])

    def _export_recursive(
        self,
        dataset: entities.Dataset = None,
        dataset_name: str = None,
        dataset_id: str = None,
        local_path: str = None,
        filters: Union[dict, entities.Filters] = None,
        annotation_filters: entities.Filters = None,
        feature_vector_filters: entities.Filters = None,
        include_feature_vectors: bool = False,
        include_annotations: bool = False,
        timeout: int = 0,
        dataset_lock: bool = False,
        lock_timeout_sec: int = None,
        export_summary: bool = False,
        max_items_per_subset: int = MAX_ITEMS_PER_SUBSET,
        export_type: ExportType = ExportType.JSON,
        output_export_type: OutputExportType = OutputExportType.JSON,
    ) -> Generator[str, None, None]:
        """
        Export dataset items recursively by splitting large datasets into smaller subsets.

        Args:
            dataset (entities.Dataset, optional): Dataset entity to export
            dataset_name (str, optional): Name of the dataset to export
            dataset_id (str, optional): ID of the dataset to export
            local_path (str, optional): Local path to save the exported data
            filters (Union[dict, entities.Filters], optional): Filters to apply on the items
            annotation_filters (entities.Filters, optional): Filters to apply on the annotations
            feature_vector_filters (entities.Filters, optional): Filters to apply on the feature vectors
            include_feature_vectors (bool, optional): Whether to include feature vectors in export. Defaults to False
            include_annotations (bool, optional): Whether to include annotations in export. Defaults to False
            timeout (int, optional): Timeout in seconds for the export operation. Defaults to 0
            dataset_lock (bool, optional): Whether to lock the dataset during export. Defaults to False
            lock_timeout_sec (int, optional): Timeout for dataset lock in seconds. Defaults to None
            export_summary (bool, optional): Whether to include export summary. Defaults to False
            max_items_per_subset (int, optional): Maximum items per subset for recursive export. Defaults to MAX_ITEMS_PER_SUBSET
            export_type (ExportType, optional): Type of export (JSON or ZIP). Defaults to ExportType.JSON
            output_export_type (OutputExportType, optional): Output format type. Defaults to OutputExportType.JSON

        Returns:
            Generator[str, None, None]: Generator yielding export paths

        Raises:
            NotImplementedError: If ZIP export type is used with JSON output type
            exceptions.PlatformException: If API request fails or command response is invalid
        """
        logger.debug(f"exporting dataset with export_type {export_type} and output_export_type {output_export_type}")
        if export_type == ExportType.ZIP and output_export_type == OutputExportType.JSON:
            raise NotImplementedError(
                "Zip export type is not supported for JSON output type.\n"
                "If Json output is required, please use the export_type = JSON"
            )

        # Get dataset entity for recursive filtering
        dataset_entity = self.get(dataset_id=self._resolve_dataset_id(dataset, dataset_name, dataset_id))
        if export_type != ExportType.JSON:
            filters_list = [filters]
        else:
            # Generate filter subsets using recursive_get_filters
            filters_list = entities.Filters._get_split_filters(
                dataset=dataset_entity, filters=filters, max_items=max_items_per_subset
            )
        # First loop: Make all API requests without waiting
        commands = []
        logger.debug("start making all API requests without waiting")
        for filter_i in filters_list:
            # Build payload for this subset
            payload = self._build_payload(
                filters=filter_i,
                include_feature_vectors=include_feature_vectors,
                include_annotations=include_annotations,
                export_type=export_type,
                annotation_filters=annotation_filters,
                feature_vector_filters=feature_vector_filters,
                dataset_lock=dataset_lock,
                lock_timeout_sec=lock_timeout_sec,
                export_summary=export_summary,
            )

            # Make API request for this subset
            success, response = self._client_api.gen_request(
                req_type='post', path=f'/datasets/{dataset_entity.id}/export', json_req=payload
            )

            if not success:
                logger.error(f"failed to make API request /datasets/{dataset_entity.id}/export with payload {payload} response {response}")
                raise exceptions.PlatformException(response)

            # Handle command execution
            commands.append( entities.Command.from_json(_json=response.json(), client_api=self._client_api))

        time.sleep(2)  # as the command have wrong progress in the beginning
        logger.debug("start waiting for all commands")
        # Second loop: Wait for all commands and process results
        for command in commands:
            command = command.wait(timeout=timeout)

            if 'outputItemId' not in command.spec:
                raise exceptions.PlatformException(
                    error='400', message="outputItemId key is missing in command response"
                )

            item_id = command.spec['outputItemId']
            # Download and process the exported item
            yield self._download_exported_item(
                item_id=item_id,
                export_type=export_type,
                local_path=local_path,
                unzip=output_export_type != OutputExportType.ZIP,
            )

    @_api_reference.add(path='/datasets/{id}/export', method='post')
    def export(
        self,
        dataset: entities.Dataset = None,
        dataset_name: str = None,
        dataset_id: str = None,
        local_path: str = None,
        filters: Union[dict, entities.Filters] = None,
        annotation_filters: entities.Filters = None,
        feature_vector_filters: entities.Filters = None,
        include_feature_vectors: bool = False,
        include_annotations: bool = False,
        export_type: ExportType = ExportType.JSON,
        timeout: int = 0,
        dataset_lock: bool = False,
        lock_timeout_sec: int = None,
        export_summary: bool = False,
        output_export_type: OutputExportType = None,
    ) -> Optional[str]:
        """
        Export dataset items and annotations.

        **Prerequisites**: You must be an *owner* or *developer* to use this method.

        You must provide at least ONE of the following params: dataset, dataset_name, dataset_id.

        **Export Behavior by Parameter Combination:**

        The behavior of this method depends on the combination of `export_type` and `output_export_type`:

        **When export_type = ExportType.JSON:**

        - **output_export_type = OutputExportType.JSON (default when None):**
          - Exports data in JSON format, split into subsets of max 500 items
          - Downloads all subset JSON files and concatenates them into a single `result.json` file
          - Returns the path to the concatenated JSON file
          - Cleans up individual subset files after concatenation

        - **output_export_type = OutputExportType.ZIP:**
          - Same as JSON export, but zips the final `result.json` file
          - Returns the path to the zipped file (`result.json.zip`)
          - Cleans up the unzipped JSON file after zipping

        - **output_export_type = OutputExportType.FOLDERS:**
          - Exports data in JSON format, split into subsets of max 500 items
          - Downloads all subset JSON files and creates individual JSON files for each item
          - Creates a folder structure mirroring the remote dataset structure
          - Returns the path to the base directory containing the folder structure
          - Each item gets its own JSON file named after the original filename

        **When export_type = ExportType.ZIP:**

        - **output_export_type = OutputExportType.ZIP:**
          - Exports data as a ZIP file containing the dataset
          - Returns the downloaded ZIP item directly
          - No additional processing or concatenation

        - **output_export_type = OutputExportType.JSON:**
          - **NOT SUPPORTED** - Raises NotImplementedError
          - Use export_type=ExportType.JSON instead for JSON output

        - **output_export_type = OutputExportType.FOLDERS:**
          - **NOT SUPPORTED** - Raises NotImplementedError
          - Use export_type=ExportType.JSON instead for folder output

        **When output_export_type = None (legacy behavior):**
        - Defaults to OutputExportType.JSON
        - Maintains backward compatibility with existing code

        :param dtlpy.entities.dataset.Dataset dataset: Dataset object
        :param str dataset_name: The name of the dataset
        :param str dataset_id: The ID of the dataset
        :param str local_path: Local path to save the exported dataset
        :param Union[dict, dtlpy.entities.filters.Filters] filters: Filters entity or a query dictionary
        :param dtlpy.entities.filters.Filters annotation_filters: Filters entity to filter annotations for export
        :param dtlpy.entities.filters.Filters feature_vector_filters: Filters entity to filter feature vectors for export
        :param bool include_feature_vectors: Include item feature vectors in the export
        :param bool include_annotations: Include item annotations in the export
        :param bool dataset_lock: Make dataset readonly during the export
        :param bool export_summary: Get Summary of the dataset export
        :param int lock_timeout_sec: Timeout for locking the dataset during export in seconds
        :param entities.ExportType export_type: Type of export ('json' or 'zip')
        :param entities.OutputExportType output_export_type: Output format ('json', 'zip', or 'folders'). If None, defaults to 'json'
        :param int timeout: Maximum time in seconds to wait for the export to complete
        :return: Path to exported file/directory, or None if export result is empty
        :rtype: Optional[str]
        """
        export_result = list(
            self._export_recursive(
                dataset=dataset,
                dataset_name=dataset_name,
                dataset_id=dataset_id,
                local_path=local_path,
                filters=filters,
                annotation_filters=annotation_filters,
                feature_vector_filters=feature_vector_filters,
                include_feature_vectors=include_feature_vectors,
                include_annotations=include_annotations,
                timeout=timeout,
                dataset_lock=dataset_lock,
                lock_timeout_sec=lock_timeout_sec,
                export_summary=export_summary,
                export_type=export_type,
                output_export_type=output_export_type,
            )
        )
        if all(x is None for x in export_result):
            logger.error("export result is empty")
            return None

        if export_type == ExportType.ZIP:
            # if export type is zip, then return the _export_recursive result as it
            return export_result[0]

        # if user didn't provide output_export_type, keep the previous behavior
        if output_export_type is None:
            output_export_type = OutputExportType.JSON

        # export type is jsos :
        # Load all items from subset JSON files and clean them up
        all_items = []
        logger.debug("start loading all items from subset JSON files")
        for json_file in export_result:
            if json_file is None:
                continue
            if os.path.isfile(json_file):
                with open(json_file, 'r') as f:
                    items = json.load(f)                    
                    if isinstance(items, list):
                        all_items.extend(items)
                os.remove(json_file)

        base_dir = os.path.dirname(export_result[0])
        if output_export_type != OutputExportType.FOLDERS:
            dataset_id=self._resolve_dataset_id(dataset, dataset_name, dataset_id)
            result_file_name = f"{dataset_id}.json"
            result_file = os.path.join(base_dir, result_file_name)
            logger.debug(f"start writing all items to result file {result_file}")
            with open(result_file, 'w') as f:
                json.dump(all_items, f)
            if output_export_type == OutputExportType.ZIP:
                # Zip the result file
                zip_filename = result_file + '.zip'
                # Create zip file
                logger.debug(f"start zipping result file {zip_filename}")
                with zipfile.ZipFile(zip_filename, 'w', zipfile.ZIP_DEFLATED) as zf:
                    zf.write(result_file, arcname=os.path.basename(result_file))

                # Remove original json after zipping
                os.remove(result_file)
                result_file = zip_filename
            return result_file
        logger.debug("start building per-item JSON files under local_path mirroring remote structure")
        # Build per-item JSON files under local_path mirroring remote structure
        for item in all_items:
            self._save_item_json_file(item_data=item, base_path=Path(base_dir), export_version=None)
        logger.debug("end building per-item JSON files under local_path mirroring remote structure")
        return base_dir

    @_api_reference.add(path='/datasets/merge', method='post')
    def merge(self,
              merge_name: str,
              dataset_ids: list,
              project_ids: str,
              with_items_annotations: bool = True,
              with_metadata: bool = True,
              with_task_annotations_status: bool = True,
              wait: bool = True):
        """
        Merge a dataset. See our `SDK docs <https://developers.dataloop.ai/tutorials/data_management/data_versioning/chapter/>`_ for more information.

        **Prerequisites**: You must be an *owner* or *developer* to use this method.

        :param str merge_name: new dataset name
        :param list dataset_ids: list id's of the datatsets you wish to merge
        :param str project_ids: the project id that include the datasets
        :param bool with_items_annotations: true to merge with items annotations
        :param bool with_metadata: true to merge with metadata
        :param bool with_task_annotations_status: true to merge with task annotations' status
        :param bool wait: wait for the command to finish
        :return: True if success
        :rtype: bool

        **Example**:

        .. code-block:: python

            success = project.datasets.merge(dataset_ids=['dataset_id1','dataset_id2'],
                                  merge_name='dataset_merge_name',
                                  with_metadata=True,
                                  with_items_annotations=False,
                                  with_task_annotations_status=False)
        """
        payload = {
            "name": merge_name,
            "datasetsIds": dataset_ids,
            "projectIds": project_ids,
            "mergeDatasetParams": {
                "withItemsAnnotations": with_items_annotations,
                "withMetadata": with_metadata,
                "withTaskAnnotationsStatus": with_task_annotations_status
            },
            'asynced': wait
        }
        success, response = self._client_api.gen_request(req_type='post',
                                                         path='/datasets/merge',
                                                         json_req=payload)

        if success:
            command = entities.Command.from_json(_json=response.json(),
                                                 client_api=self._client_api)
            if not wait:
                return command
            command = command.wait(timeout=0)
            if 'mergeDatasetsConfiguration' not in command.spec:
                raise exceptions.PlatformException(error='400',
                                                   message="mergeDatasetsConfiguration key is missing in command response: {}"
                                                   .format(response))
            return True
        else:
            raise exceptions.PlatformException(response)

    @_api_reference.add(path='/datasets/{id}/sync', method='post')
    def sync(self, dataset_id: str, wait: bool = True):
        """
        Sync dataset with external storage.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param str dataset_id: The Id of the dataset to sync
        :param bool wait: wait for the command to finish
        :return: True if success
        :rtype: bool

        **Example**:

        .. code-block:: python

            success = project.datasets.sync(dataset_id='dataset_id')
        """

        success, response = self._client_api.gen_request(req_type='post',
                                                         path='/datasets/{}/sync'.format(dataset_id))

        if success:
            command = entities.Command.from_json(_json=response.json(),
                                                 client_api=self._client_api)
            if not wait:
                return command
            command = command.wait(timeout=0)
            if 'datasetId' not in command.spec:
                raise exceptions.PlatformException(error='400',
                                                   message="datasetId key is missing in command response: {}"
                                                   .format(response))
            return True
        else:
            raise exceptions.PlatformException(response)

    @_api_reference.add(path='/datasets', method='post')
    def create(self,
               dataset_name: str,
               labels=None,
               attributes=None,
               ontology_ids=None,
               driver: entities.Driver = None,
               driver_id: str = None,
               checkout: bool = False,
               expiration_options: entities.ExpirationOptions = None,
               index_driver: entities.IndexDriver = None,
               recipe_id: str = None
               ) -> entities.Dataset:
        """
        Create a new dataset

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param str dataset_name: The Name of the dataset
        :param list labels: dictionary of {tag: color} or list of label entities
        :param list attributes: dataset's ontology's attributes
        :param list ontology_ids: optional - dataset ontology
        :param dtlpy.entities.driver.Driver driver: optional - storage driver Driver object or driver name
        :param str driver_id: optional - driver id
        :param bool checkout: set the dataset as a default dataset object (cookies)
        :param ExpirationOptions expiration_options: dl.ExpirationOptions object that contain definitions for dataset like MaxItemDays
        :param str index_driver: dl.IndexDriver, dataset driver version
        :param str recipe_id: optional - recipe id
        :return: Dataset object
        :rtype: dtlpy.entities.dataset.Dataset

        **Example**:

        .. code-block:: python

            dataset = project.datasets.create(dataset_name='dataset_name', ontology_ids='ontology_ids')
        """
        create_default_recipe = True
        if any([labels, attributes, ontology_ids, recipe_id]):
            create_default_recipe = False

        # labels to list
        if labels is not None:
            if not isinstance(labels, list):
                labels = [labels]
            if not all(isinstance(label, entities.Label) for label in labels):
                labels = entities.Dataset.serialize_labels(labels)
        else:
            labels = list()

        # get creator from token
        payload = {'name': dataset_name,
                   'projects': [self.project.id],
                   'createDefaultRecipe': create_default_recipe
                   }

        if driver_id is None and driver is not None:
            if isinstance(driver, entities.Driver):
                driver_id = driver.id
            elif isinstance(driver, str):
                driver_id = self.project.drivers.get(driver_name=driver).id
            else:
                raise exceptions.PlatformException(
                    error=400,
                    message='Input arg "driver" must be Driver object or a string driver name. got type: {!r}'.format(
                        type(driver)))
        if driver_id is not None:
            payload['driver'] = driver_id

        if expiration_options:
            payload['expirationOptions'] = expiration_options.to_json()
        if index_driver is not None:
            payload['indexDriver'] = index_driver

        success, response = self._client_api.gen_request(req_type='post',
                                                         path='/datasets',
                                                         json_req=payload)
        if success:
            dataset = entities.Dataset.from_json(client_api=self._client_api,
                                                 _json=response.json(),
                                                 datasets=self,
                                                 project=self.project)
            # create ontology and recipe
            if not create_default_recipe:
                if recipe_id is not None:
                    dataset.switch_recipe(recipe_id=recipe_id)
                else:
                    dataset = dataset.recipes.create(ontology_ids=ontology_ids,
                                                     labels=labels,
                                                     attributes=attributes).dataset
        else:
            raise exceptions.PlatformException(response)
        logger.info('Dataset was created successfully. Dataset id: {!r}'.format(dataset.id))
        assert isinstance(dataset, entities.Dataset)
        if checkout:
            self.checkout(dataset=dataset)
        return dataset

    @staticmethod
    def _convert_single(downloader,
                        item,
                        img_filepath,
                        local_path,
                        overwrite,
                        annotation_options,
                        annotation_filters,
                        thickness,
                        with_text,
                        progress,
                        alpha,
                        export_version):
        # this is to convert the downloaded json files to any other annotation type
        try:
            if entities.ViewAnnotationOptions.ANNOTATION_ON_IMAGE in annotation_options:
                if img_filepath is None:
                    img_filepath = item.download()
            downloader._download_img_annotations(item=item,
                                                 img_filepath=img_filepath,
                                                 local_path=local_path,
                                                 overwrite=overwrite,
                                                 annotation_options=annotation_options,
                                                 annotation_filters=annotation_filters,
                                                 thickness=thickness,
                                                 alpha=alpha,
                                                 with_text=with_text,
                                                 export_version=export_version
                                                 )
        except Exception:
            logger.error('Failed to download annotation for item: {!r}'.format(item.name))
        progress.update()

    @staticmethod
    def download_annotations(dataset: entities.Dataset,
                             local_path: str = None,
                             filters: entities.Filters = None,
                             annotation_options: entities.ViewAnnotationOptions = None,
                             annotation_filters: entities.Filters = None,
                             overwrite: bool = False,
                             thickness: int = 1,
                             with_text: bool = False,
                             remote_path: str = None,
                             include_annotations_in_output: bool = True,
                             export_png_files: bool = False,
                             filter_output_annotations: bool = False,
                             alpha: float = 1,
                             export_version=entities.ExportVersion.V1,
                             dataset_lock: bool = False, 
                             lock_timeout_sec: int = None,
                             export_summary: bool = False,              
                             ) -> str:
        """
        Download dataset's annotations by filters.

        You may filter the dataset both for items and for annotations and download annotations.

        Optional -- download annotations as: mask, instance, image mask of the item.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param dtlpy.entities.dataset.Dataset dataset: dataset object
        :param str local_path: local folder or filename to save to.
        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :param list annotation_options: type of download annotations: list(dl.ViewAnnotationOptions)
        :param dtlpy.entities.filters.Filters annotation_filters: Filters entity to filter annotations for download
        :param bool overwrite: optional - default = False to overwrite the existing files
        :param bool dataset_loc: optional - default = False to make the dataset readonly
        :param int thickness: optional - line thickness, if -1 annotation will be filled, default =1
        :param bool with_text: optional - add text to annotations, default = False
        :param str remote_path: DEPRECATED and ignored
        :param bool include_annotations_in_output: default - False , if export should contain annotations
        :param bool export_png_files: default - if True, semantic annotations should be exported as png files
        :param bool filter_output_annotations: default - False, given an export by filter - determine if to filter out annotations
        :param float alpha: opacity value [0 1], default 1
        :param str export_version:  exported items will have original extension in filename, `V1` - no original extension in filenames
        :return: local_path of the directory where all the downloaded item
        :param bool dataset_lock: optional - default = False
        :param bool export_summary: optional - default = False
        :param int lock_timeout_sec: optional
        :rtype: str

        **Example**:

        .. code-block:: python

            file_path = project.datasets.download_annotations(dataset='dataset_entity',
                                                 local_path='local_path',
                                                 annotation_options=dl.ViewAnnotationOptions,
                                                 overwrite=False,
                                                 thickness=1,
                                                 with_text=False,
                                                 alpha=1,
                                                 dataset_lock=False,
                                                 lock_timeout_sec=300,
                                                 export_summary=False                                               
                                                 )
        """
        if annotation_options is None:
            annotation_options = list()
        elif not isinstance(annotation_options, list):
            annotation_options = [annotation_options]
        for ann_option in annotation_options:
            if ann_option not in entities.ViewAnnotationOptions:
                raise PlatformException(
                    error='400',
                    message=f'Unknown annotation download option: {ann_option}, please choose from: {list(entities.ViewAnnotationOptions)}',
                )
        if remote_path is not None:
            logger.warning(f'"remote_path" is ignored. Use "filters=dl.Filters(field="dir, values={remote_path!r}"')
        if filter_output_annotations is True:
            logger.warning("'filter_output_annotations' is ignored but kept for legacy support")
        if include_annotations_in_output is False:
            logger.warning("include_annotations_in_output was False, but was set to True since this function downloads annotations.")
            include_annotations_in_output = True

        if local_path is None:
            if dataset.project is None:
                # by dataset name
                local_path = str(Path(service_defaults.DATALOOP_PATH) / "datasets" / f"{dataset.name}_{dataset.id}")
            else:
                # by dataset and project name
                local_path = str(Path(service_defaults.DATALOOP_PATH) / "projects" / dataset.project.name / "datasets" / dataset.name)

        if filters is None:
            filters = entities.Filters()
            filters._user_query = 'false'
        if annotation_filters is not None:
            for annotation_filter_and in annotation_filters.and_filter_list:
                filters.add_join(field=annotation_filter_and.field,
                                 values=annotation_filter_and.values,
                                 operator=annotation_filter_and.operator,
                                 method=entities.FiltersMethod.AND)
            for annotation_filter_or in annotation_filters.or_filter_list:
                filters.add_join(field=annotation_filter_or.field,
                                 values=annotation_filter_or.values,
                                 operator=annotation_filter_or.operator,
                                 method=entities.FiltersMethod.OR)

        downloader = repositories.Downloader(items_repository=dataset.items)

        # Setup for incremental processing
        if len(annotation_options) == 0 :
            pool = None
            progress = None
            jobs = []
        else:             
            # Get total count for progress bar
            filter_copy = copy.deepcopy(filters)
            filter_copy.page_size = 0
            pages = dataset.items.list(filters=filter_copy)
            total_items = pages.items_count

            # Setup thread pool and progress bar
            pool = dataset._client_api.thread_pools(pool_name='dataset.download')
            progress = tqdm.tqdm(
                total=total_items,
                disable=dataset._client_api.verbose.disable_progress_bar_download_annotations,
                file=sys.stdout,
                desc='Download Annotations'
            )
            jobs = []
            

        # Call _export_recursive as generator
        export_generator = dataset.project.datasets._export_recursive(
            dataset=dataset,
            local_path=tempfile.mkdtemp(prefix='annotations_jsons_'),
            filters=filters,
            annotation_filters=annotation_filters,
            include_annotations=True,
            export_type=ExportType.JSON,
            dataset_lock=dataset_lock,
            lock_timeout_sec=lock_timeout_sec,
            export_summary=export_summary,
            timeout=0,
            max_items_per_subset=DOWNLOAD_ANNOTATIONS_MAX_ITEMS_PER_SUBSET
        )

        # Process each subset JSON file incrementally
        for subset_json_file in export_generator:
            if subset_json_file is None or not Path(subset_json_file).is_file():
                continue

            try:
                # Open and load the items array
                with open(subset_json_file, 'r') as f:
                    items_data = json.load(f)

                # Process each item immediately
                for item_data in items_data:
                    # Split and save individual JSON file
                    Datasets._save_item_json_file(item_data=item_data, base_path=Path(local_path) / 'json', export_version=export_version)

                    # If annotation_options are provided, submit to thread pool immediately
                    if annotation_options:
                        # Create Item entity from item_data
                        item = entities.Item.from_json(
                            _json=item_data,
                            client_api=dataset._client_api,
                            dataset=dataset
                        )

                        job = pool.submit(
                            Datasets._convert_single,
                            **{
                                'downloader': downloader,
                                'item': item,
                                'img_filepath': None,
                                'local_path': local_path,
                                'overwrite': overwrite,
                                'annotation_options': annotation_options,
                                'annotation_filters': annotation_filters,
                                'thickness': thickness,
                                'with_text': with_text,
                                'progress': progress,
                                'alpha': alpha,
                                'export_version': export_version
                            }
                        )
                        jobs.append(job)

                # Clean up temporary subset JSON file
                os.remove(subset_json_file)
            except Exception as e:
                logger.exception(f'Failed processing subset JSON file {subset_json_file}: {e}')

        # Wait for all thread pool jobs to complete
        if annotation_options:
            _ = [j.result() for j in jobs]
            progress.close()

        return local_path

    def _upload_single_item_annotation(self, item, file, pbar):
        try:
            item.annotations.upload(file)
        except Exception as err:
            raise err
        finally:
            pbar.update()

    def upload_annotations(self,
                           dataset,
                           local_path,
                           filters: entities.Filters = None,
                           clean=False,
                           remote_root_path='/',
                           export_version=entities.ExportVersion.V1
                           ):
        """
        Upload annotations to dataset. 

        Example for remote_root_path: If the item filepath is "/a/b/item" and remote_root_path is "/a" - the start folder will be b instead of a

        **Prerequisites**: You must have a dataset with items that are related to the annotations. The relationship between the dataset and annotations is shown in the name. You must be in the role of an *owner* or *developer*.

        :param dtlpy.entities.dataset.Dataset dataset: dataset to upload to
        :param str local_path: str - local folder where the annotations files are
        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :param bool clean: True to remove the old annotations
        :param str remote_root_path: the remote root path to match remote and local items
        :param str export_version:  exported items will have original extension in filename, `V1` - no original extension in filenames

        **Example**:

        .. code-block:: python

            project.datasets.upload_annotations(dataset='dataset_entity',
                                                 local_path='local_path',
                                                 clean=False,
                                                 export_version=dl.ExportVersion.V1
                                                 )
        """
        if filters is None:
            filters = entities.Filters()
            filters._user_query = 'false'
        pages = dataset.items.list(filters=filters)
        total_items = pages.items_count
        pbar = tqdm.tqdm(total=total_items, disable=dataset._client_api.verbose.disable_progress_bar_upload_annotations,
                         file=sys.stdout, desc='Upload Annotations')
        pool = self._client_api.thread_pools('annotation.upload')
        annotations_uploaded_count = 0
        for item in pages.all():
            if export_version == entities.ExportVersion.V1:
                _, ext = os.path.splitext(item.filename)
                filepath = item.filename.replace(ext, '.json')
            else:
                filepath = item.filename + '.json'
            # make the file path ignore the hierarchy of the files that in remote_root_path
            filepath = os.path.relpath(filepath, remote_root_path)
            json_file = os.path.join(local_path, filepath)
            if not os.path.isfile(json_file):
                pbar.update()
                continue
            annotations_uploaded_count += 1
            if item.annotated and clean:
                item.annotations.delete(filters=entities.Filters(resource=entities.FiltersResource.ANNOTATION))
            pool.submit(self._upload_single_item_annotation, **{'item': item,
                                                                'file': json_file,
                                                                'pbar': pbar})
        pool.shutdown()
        if annotations_uploaded_count == 0:
            logger.warning(msg="No annotations uploaded to dataset! ")
        else:
            logger.info(msg='Found and uploaded {} annotations.'.format(annotations_uploaded_count))

    def set_readonly(self, state: bool, dataset: entities.Dataset):
        """
        Set dataset readonly mode.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param bool state: state to update readonly mode
        :param dtlpy.entities.dataset.Dataset dataset: dataset object

        **Example**:

        .. code-block:: python

            project.datasets.set_readonly(dataset='dataset_entity', state=True)
        """
        import warnings
        warnings.warn("`readonly` flag on dataset is deprecated, doing nothing.", DeprecationWarning)

    @_api_reference.add(path='/datasets/{id}/split', method='post')
    def split_ml_subsets(self,
                        dataset_id: str,
                        items_query: entities.filters,
                        ml_split_list: dict) -> bool:
        """
        Split dataset items into ML subsets.

        :param str dataset_id: The ID of the dataset.
        :param dict items_query: Query to select items.
        :param dict ml_split_list: Dictionary with 'train', 'validation', 'test' keys and integer percentages.
        :return: True if the split operation was successful.
        :rtype: bool
        :raises: PlatformException on failure and ValueError if percentages do not sum to 100 or invalid keys/values.
        """
        # Validate percentages
        if not ml_split_list:
            ml_split_list = {'train': 80, 'validation': 10, 'test': 10}

        if not items_query:
            items_query = entities.Filters()

        items_query_dict = items_query.prepare()
        required_keys = {'train', 'validation', 'test'}
        if set(ml_split_list.keys()) != required_keys:
            raise ValueError("MLSplitList must have exactly the keys 'train', 'validation', 'test'.")
        total = sum(ml_split_list.values())
        if total != 100:
            raise ValueError(
                "Please set the Train, Validation, and Test subsets percentages to add up to 100%. "
                "For example: 70, 15, 15."
            )
        for key, value in ml_split_list.items():
            if not isinstance(value, int) or value < 0:
                raise ValueError("Percentages must be integers >= 0.")
        payload = {
            'itemsQuery': items_query_dict,
            'MLSplitList': ml_split_list
        }
        path = f'/datasets/{dataset_id}/split'
        success, response = self._client_api.gen_request(req_type='post',
                                                        path=path,
                                                        json_req=payload)
        if success:
            # Wait for the split operation to complete
            command = entities.Command.from_json(_json=response.json(),
                                                client_api=self._client_api)
            command.wait()
            return True
        else:
            raise exceptions.PlatformException(response)

    @_api_reference.add(path='/datasets/{id}/items/bulk-update-metadata', method='post')
    def bulk_update_ml_subset(self, dataset_id: str, items_query: dict, subset: str = None, deleteTag: bool = False) -> bool:
        """
        Bulk update ML subset assignment for selected items.
        If subset is None, remove subsets. Otherwise, assign the specified subset.

        :param str dataset_id: ID of the dataset
        :param dict items_query: DQLResourceQuery (filters) for selecting items
        :param str subset: 'train', 'validation', 'test' or None to remove all
        :return: True if success
        :rtype: bool
        """
        if items_query is None:
            items_query = entities.Filters()
        items_query_dict = items_query.prepare()
        if not deleteTag and subset not in ['train', 'validation', 'test']:
            raise ValueError("subset must be one of: 'train', 'validation', 'test'")
        # Determine tag values based on subset
        tags = {
            'train': True if subset == 'train' else None,
            'validation': True if subset == 'validation' else None,
            'test': True if subset == 'test' else None
        }

        payload = {
            "query": items_query_dict,
            "updateQuery": {
                "update": {
                    "metadata": {
                        "system": {
                            "tags": tags
                        }
                    }
                },
                "systemSpace": True
            }
        }

        success, response = self._client_api.gen_request(
            req_type='post',
            path=f'/datasets/{dataset_id}/items/bulk-update-metadata',
            json_req=payload
        )
        if success:
            # Similar to split operation, a command is returned
            command = entities.Command.from_json(_json=response.json(), client_api=self._client_api)
            command.wait()
            return True
        else:
            raise exceptions.PlatformException(response)



"""
Integrations Repository
"""
import base64
import json
import logging
from .. import entities, exceptions, miscellaneous, _api_reference
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')


class Integrations:
    """
    Integrations Repository

    The Integrations class allows you to manage data integrations from your external storage (e.g., S3, GCS, Azure)
    into your Dataloop's Dataset storage, as well as sync data in your Dataloop's Datasets with data in your external
    storage.

    For more information on Organization Storage Integration see the `Dataloop documentation <https://dataloop.ai/docs/organization-integrations>`_  and `developers' docs <https://developers.dataloop.ai/tutorials/data_management/>`_.

    """

    def __init__(self, client_api: ApiClient, org: entities.Organization = None,
                 project: entities.Project = None):
        self._client_api = client_api
        self._org = org
        self._project = project

    @property
    def project(self) -> entities.Project:
        return self._project

    @project.setter
    def project(self, project: entities.Project):
        if not isinstance(project, entities.Project):
            raise ValueError('Must input a valid Project entity')
        self._project = project

    @property
    def org(self) -> entities.Organization:
        if self._org is None:
            if self.project is not None:
                self._org = entities.Organization.from_json(_json=self.project.org, client_api=self._client_api)
        return self._org

    @org.setter
    def org(self, org: entities.Organization):
        if not isinstance(org, entities.Organization):
            raise ValueError('Must input a valid Organization entity')
        self._org = org

    @_api_reference.add(path='/orgs/{orgId}/integrations/{integrationId}', method='delete')
    def delete(self,
               integrations_id: str,
               sure: bool = False,
               really: bool = False,
               organization_id: str = None
               ) -> bool:
        """
        Delete integrations from the organization.

        **Prerequisites**: You must be an organization *owner* to delete an integration.

        :param organization_id: organization id
        :param str integrations_id: integrations id
        :param bool sure: Are you sure you want to delete?
        :param bool really: Really really sure?
        :return: success
        :rtype: bool

        **Example**:

        .. code-block:: python

            project.integrations.delete(integrations_id='integrations_id', sure=True, really=True)
        """
        if sure and really:
            if self.project is None and self.org is None and organization_id is None:
                raise exceptions.PlatformException(
                    error='400',
                    message='Must provide an identifier in inputs')

        if organization_id is None:
            if self.project is not None:
                organization_id = self.project.org.get('id')
            else:
                organization_id = self.org.id

            url_path = '/orgs/{}/integrations/{}'.format(organization_id, integrations_id)
            success, response = self._client_api.gen_request(req_type='delete',
                                                             path=url_path)
            if not success:
                raise exceptions.PlatformException(response)
            else:
                return True
        else:
            raise exceptions.PlatformException(
                error='403',
                message='Cant delete integrations from SDK. Please login to platform to delete')

    @_api_reference.add(path='/orgs/{orgId}/integrations', method='post')
    def create(self,
               integrations_type: entities.IntegrationType,
               name: str,
               options: dict,
               metadata: dict = None,
               organization_id: str = None,
               ):
        """
        Create an integration between an external storage and the organization.

        **Examples for options include**:
        s3 - {key: "", secret: ""};
        gcs - {key: "", secret: "", content: ""};
        azureblob - {key: "", secret: "", clientId: "", tenantId: ""};
        key_value - {key: "", value: ""}
        aws-sts - {key: "", secret: "", roleArns: ""}
        aws-cross - {}
        gcp-cross - {}
        gcp-workload-identity-federation - {"secret": "", "content": "{}", "clientId": ""}
        private-registry (ECR) - can use generate_ecr_options to generate the options
        private-registry (GAR) - use generate_gar_options to generate the options
        private-registry (ACR) - use generate_azure_container_registry_options to generate the options
        private-registry (DockerHub) - use generate_docker_hub_options to generate the options

        **Prerequisites**: You must be an *owner* in the organization.

        :param IntegrationType integrations_type: integrations type dl.IntegrationType
        :param str name: integrations name
        :param dict options: dict of storage secrets
        :param dict metadata: metadata
        :param str organization_id: organization id
        :return: success
        :rtype: bool

        **Example**:

        .. code-block:: python

            project.integrations.create(integrations_type=dl.IntegrationType.S3,
                            name='S3Integration',
                            options={key: "Access key ID", secret: "Secret access key"})
        """

        if self.project is None and self.org is None and organization_id is None:
            raise exceptions.PlatformException(
                error='400',
                message='Must have an organization or project')

        if organization_id is None:
            if self.project is not None:
                organization_id = self.project.org.get('id')
            else:
                organization_id = self.org.id

        url_path = '/orgs/{}/integrations'.format(organization_id)
        payload = {"type": integrations_type.value if isinstance(integrations_type,
                                                                 entities.IntegrationType) else integrations_type,
                   'name': name, 'options': options}
        if metadata is not None:
            payload['metadata'] = metadata
        success, response = self._client_api.gen_request(req_type='post',
                                                         path=url_path,
                                                         json_req=payload)
        if not success:
            raise exceptions.PlatformException(response)
        else:
            integration = entities.Integration.from_json(_json=response.json(), client_api=self._client_api)
        if integration.metadata and isinstance(integration.metadata, list) and len(integration.metadata) > 0:
            for m in integration.metadata:
                if m['name'] == 'status':
                    integration_status = m['value']
                    logger.info('Integration status: {}'.format(integration_status))
        return integration

    @_api_reference.add(path='/orgs/{orgId}/integrations', method='patch')
    def update(self,
               new_name: str = None,
               integrations_id: str = None,
               integration: entities.Integration = None,
               new_options: dict = None,
               organization_id: str = None,
               reload_services: bool = None,
               ):
        """
        Update the integration's name.

        **Prerequisites**: You must be an *owner* in the organization.

        :param str new_name: new name
        :param str integrations_id: integrations id
        :param Integration integration: integration object
        :param dict new_options: new value
        :param str organization_id: organization id
        :param bool reload_services: reload services associated with this integration
        :return: Integration object
        :rtype: dtlpy.entities.integration.Integration

        **Examples for options include**:
        s3 - {key: "", secret: ""};
        gcs - {key: "", secret: "", content: ""};
        azureblob - {key: "", secret: "", clientId: "", tenantId: ""};
        key_value - {key: "", value: ""}
        aws-sts - {key: "", secret: "", roleArns: ""}
        aws-cross - {roleArn: ""}
        gcp-cross - {"email: "", "resourceName": ""}

        **Example**:

        .. code-block:: python

            project.integrations.update(integrations_id='integrations_id', new_options={roleArn: ""})
        """

        if self.project is None and self.org is None and organization_id is None:
            raise exceptions.PlatformException(
                error='400',
                message='Must have an organization or project')
        if integrations_id is None and integration is None:
            raise exceptions.PlatformException(
                error='400',
                message='Must have an integrations_id or integration')

        if organization_id is None:
            if self.project is not None:
                organization_id = self.project.org.get('id')
            else:
                organization_id = self.org.id

        if reload_services is None:
            logger.warning(
                "Param reload_services was not provided. If the integration you are updating is used\n"
                "in FaaS services these services will keep using the old value until updated."
            )

        url_path = '/orgs/{org_id}/integrations{query_params}'.format(
            org_id=organization_id,
            query_params='?reloadServices=true' if reload_services else ''
        )
        payload = dict(integrationId=integrations_id if integrations_id is not None else integration.id)
        if new_name is not None:
            payload['name'] = new_name
        if new_options is not None:
            if integration is None:
                integration = self.get(integrations_id=integrations_id)
            payload['credentials'] = dict(options=new_options, type=integration.type)

        success, response = self._client_api.gen_request(req_type='patch',
                                                         path=url_path,
                                                         json_req=payload)
        if not success:
            raise exceptions.PlatformException(response)

        return entities.Integration.from_json(_json=response.json(), client_api=self._client_api)

    @_api_reference.add(path='/orgs/{orgId}/integrations/{integrationId}', method='get')
    def get(self, integrations_id: str, organization_id: str = None):
        """
        Get organization integrations. Use this method to access your integration and be able to use it in your code.

        **Prerequisites**: You must be an *owner* in the organization.

        :param str integrations_id: integrations id
        :param str organization_id: organization id
        :return: Integration object
        :rtype: dtlpy.entities.integration.Integration

        **Example**:

        .. code-block:: python

            project.integrations.get(integrations_id='integrations_id')
        """
        if self.project is None and self.org is None and organization_id is None:
            raise exceptions.PlatformException(
                error='400',
                message='Must have an organization or project')

        if organization_id is None:
            if self.project is not None:
                organization_id = self.project.org.get('id')
            else:
                organization_id = self.org.id

        url_path = '/orgs/{}/integrations/{}'.format(organization_id, integrations_id)

        success, response = self._client_api.gen_request(req_type='get',
                                                         path=url_path)
        if not success:
            raise exceptions.PlatformException(response)
        return entities.Integration.from_json(_json=response.json(), client_api=self._client_api)

    @_api_reference.add(path='/orgs/{orgId}/integrations', method='get')
    def list(self, only_available=False, organization_id: str = None):
        """
        List all the organization's integrations with external storage.

        **Prerequisites**: You must be an *owner* in the organization.

        :param bool only_available: if True list only the available integrations.
        :param str organization_id: organization id
        :return: groups list
        :rtype: list

        **Example**:

        .. code-block:: python

            project.integrations.list(only_available=True)
        """
        if self.project is None and self.org is None and organization_id is None:
            raise exceptions.PlatformException(
                error='400',
                message='Must have an organization or project')

        if organization_id is None:
            if self.project is not None:
                organization_id = self.project.org.get('id')
            else:
                organization_id = self.org.id

        if only_available:
            url_path = '/orgs/{}/availableIntegrations'.format(organization_id)
        else:
            url_path = '/orgs/{}/integrations'.format(organization_id)

        success, response = self._client_api.gen_request(req_type='get',
                                                         path=url_path)
        if not success:
            raise exceptions.PlatformException(response)

        available_integrations = miscellaneous.List(response.json())
        return available_integrations

    @staticmethod
    def generate_gar_options(service_account: str, location: str, email: str = None) -> dict:
        """
        Generates a Google Artifact Registry JSON configuration and returns it as a base64-encoded string.

        Parameters:
            location (str): The region where the repository will be created (e.g., 'us-central1').
            service_account (str): The service_account parameter represents the Google Cloud service account credentials
                                    in the form of a JSON key file. This JSON contains the private key and other metadata
                                    required for authenticating with Google Artifact Registry. It is used to generate a Kubernetes secret
                                    that stores the credentials for pulling container images from the registry.
                                    The JSON key must include fields such as client_email, private_key, and project_id,
                                    and it is typically downloaded from the Google Cloud Console when creating the service account

        Returns:
            str: A base64-encoded string representation of the repository JSON configuration.
        """
        return IntegrationUtils.generate_gar_options(service_account=service_account, location=location, email=email)

    @staticmethod
    def generate_docker_hub_options(username: str, password: str, email: str = None) -> dict:
        """
        Generates a Docker Hub JSON configuration and returns it as a base64-encoded string.

        Parameters:
            username (str): The Docker Hub username.
            password (str): The Docker Hub password.
            email (str): Optional - Docker Hub email.

        Returns:
            str: A base64-encoded string representation of the repository JSON configuration.
        """
        return IntegrationUtils.generate_docker_hub_options(username=username, password=password, email=email)

    @staticmethod
    def generate_azure_container_registry_options(username: str, password: str, location: str) -> dict:
        """
        Generates an Azure Container Registry JSON configuration and returns it as a base64-encoded string.

        Parameters:
            username (str): The Azure username.
            password (str): The Azure password.
            location (str): server URL of Azure Container Registry

        Returns:
            str: A base64-encoded string representation of the repository JSON configuration.
        """
        return IntegrationUtils.generate_docker_hub_options(username=username, password=password, location=location)

    @staticmethod
    def generate_ecr_options(access_key_id: str, secret_access_key: str, account: str, region: str) -> dict:
        """
        Generates an Amazon Elastic Container Registry (ECR) JSON configuration and returns it as a base64-encoded string.

        Parameters:
            access_key_id (str): The AWS access key ID.
            secret_access_key (str): The AWS secret access key.
            account (str): The AWS account ID.
            region (str): The AWS region.

        Returns:
            str: A base64-encoded string representation of the repository JSON configuration.
        """
        return IntegrationUtils.generate_ecr_options(
            access_key_id=access_key_id,
            secret_access_key=secret_access_key,
            account=account,
            region=region
        )


class IntegrationUtils:

    @staticmethod
    def encode(st: str):
        return str(base64.b64encode(bytes(st, 'utf-8')))[2:-1]

    @staticmethod
    def generate_json_key_options(location: str, username: str, password: str, auth: str, email: str = None):
        encoded_pass = {
            "auths": {
                f"{location}": {
                    "username": username,
                    "password": password,
                    "auth": auth
                }
            }
        }

        if email:
            encoded_pass['auths'][f'{location}']['email'] = email

        return {
            "name": "_json_key",
            "spec": {
                "password": IntegrationUtils.encode(json.dumps(encoded_pass))
            }
        }

    @staticmethod
    def generate_gar_options(service_account: str, location: str, email: str = None) -> dict:

        if not service_account:
            raise ValueError('Missing Service Account')
        if not location:
            raise ValueError('Missing Location')

        username = "_json_key"
        cred = f"{username}:{service_account}"
        auth = IntegrationUtils.encode(cred)

        return IntegrationUtils.generate_json_key_options(
            location=location,
            username=username,
            password=service_account,
            auth=auth,
            email=email
        )

    @staticmethod
    def generate_docker_hub_options(username: str, password: str, email: str = None, location='docker.io') -> dict:

        if not username:
            raise ValueError('Missing Username')
        if not password:
            raise ValueError('Missing Password')

        auth = IntegrationUtils.encode('{}:{}'.format(username, password))

        return IntegrationUtils.generate_json_key_options(
            location=location,
            username=username,
            password=password,
            auth=auth,
            email=email
        )

    @staticmethod
    def generate_ecr_options(access_key_id: str, secret_access_key: str, account: str, region: str) -> dict:
        return {
            "name": "AWS",
            "spec": {
                "accessKeyId": access_key_id,
                "secretAccessKey": secret_access_key,
                "account": account,
                "region": region,
            }
        }



import time
import logging
import threading
import traceback

logger = logging.getLogger(name='dtlpy')


def check_in_thread(version, client_api):
    try:
        # check for a valid token
        if client_api.token_expired():
            # wait for user to maybe login in the next 2 minutes
            time.sleep(120)
        # check for a valid token again
        if client_api.token_expired():
            # return if cant find a valid token
            logger.debug('Cant check_sdk without a valid token.')
            return

        # try read token for email
        try:
            user_email = client_api.info()['user_email']
        except Exception:
            user_email = 'na'
        logger.debug('SDK info: user: {}, version: {}'.format(user_email, version))
        return_type, resp = client_api.gen_request(req_type='POST',
                                                   path='/sdk/check',
                                                   data={'version': version,
                                                         'email': user_email},
                                                   log_error=False)
        if resp.ok:
            resp = resp.json()
            client_api.cookie_io.put(key='check_version_status',
                                     value={'level': resp['level'],
                                            'msg': resp['msg']})
        else:
            client_api.cookie_io.put(key='check_version_status',
                                     value={'level': 'debug',
                                            'msg': 'unknown'})
    except Exception:
        logger.debug(traceback.format_exc())
        logger.debug('Error in check sdk manager.')


def check(version, client_api):
    worker = threading.Thread(target=check_in_thread,
                              kwargs={'version': version,
                                      'client_api': client_api})
    worker.daemon = True
    worker.start()
    status = client_api.cookie_io.get('check_version_status')
    if status is not None:
        level = status['level']
        msg = status['msg']
        if level.lower() == 'debug':
            logger.debug(msg=msg)
        elif level.lower() == 'info':
            logger.info(msg=msg)
        elif level.lower() == 'warning':
            logger.warning(msg=msg)
        elif level.lower() == 'error':
            logger.error(msg=msg)
        else:
            logger.debug(msg='unknown')



from . import BaseAnnotationDefinition


class Comparison(BaseAnnotationDefinition):
    """
    Comparison annotation object
    """
    type = "comparison"

    def __init__(self, coordinates, label, attributes=None, description=None):
        super().__init__(description=description, attributes=attributes)
        self.label = label
        self.coordinates = coordinates

    @property
    def geo(self):
        return 0

    @property
    def left(self):
        return 0

    @property
    def top(self):
        return 0

    @property
    def right(self):
        return 0

    @property
    def bottom(self):
        return 0

    @property
    def z(self):
        return 0

    def show(self, image, thickness, with_text, height, width, annotation_format, color, alpha=1):
        """
        Show annotation as ndarray
        :param image: empty or image to draw on
        :param thickness:
        :param with_text: not required
        :param height: item height
        :param width: item width
        :param annotation_format: options: list(dl.ViewAnnotationOptions)
        :param color: color
        :param alpha: opacity value [0 1], default 1
        :return: ndarray
        """
        if with_text:
            image = self.add_text_to_image(image=image, annotation=self)
        return image

    def to_coordinates(self, color):
        return self.coordinates

    @classmethod
    def from_json(cls, _json):
        if "coordinates" in _json:
            coordinates = _json["coordinates"]
        elif "data" in _json:
            coordinates = _json["data"]
        else:
            coordinates = dict()

        return cls(
            coordinates=coordinates,
            label=_json.get("label", None),
            attributes=_json.get("attributes", None)
        )



import dtlpy as dl
import logging
import datetime

logger = logging.getLogger(name='dtlpy')


class ServiceRunner(dl.BaseServiceRunner):
    """
    Package runner class

    """

    def __init__(self):
        """
        Init package attributes here

        :return:
        """

    def run(self, dataset, progress=None):
        """
        In this example we create a task for annotated items in dataset

        :param progress: Use this to update the progress of your package
        :return:
        """
        # these lines can be removed
        assert isinstance(progress, dl.Progress)
        assert isinstance(dataset, dl.entities.Dataset)
        filters = dl.Filters(field='annotated', values=True)
        filters.add_join(field='label', values='person')
        task = dataset.tasks.create(task_name='AutomatedTask',
                                    due_date=datetime.datetime.now().timestamp() + 60*60*24*7,
                                    assignee_ids=['annotator1@dataloop.ai', 'annotator2@dataloop.ai'])
        logger.info('Task created successfully. Task name: {}'.format(task.name))


if __name__ == "__main__":
    """
    Run this main to locally debug your package
    """
    dl.packages.test_local_package()



from collections import namedtuple
import traceback
import logging
from enum import Enum

import attr
import os

from .. import repositories, entities, services, exceptions
from ..services.api_client import ApiClient
from .annotation import ViewAnnotationOptions, AnnotationType, ExportVersion

logger = logging.getLogger(name='dtlpy')


class IndexDriver(str, Enum):
    V1 = "v1"
    V2 = "v2"


class ExportType(str, Enum):
    JSON = "json"
    ZIP = "zip"

class OutputExportType(str, Enum):
    JSON = "json"
    ZIP = "zip"
    FOLDERS = "folders"

class ExpirationOptions:
    """
    ExpirationOptions object
    """

    def __init__(self, item_max_days: int = None):
        """
        :param item_max_days: int. items in dataset will be auto delete after this number id days
        """
        self.item_max_days = item_max_days

    def to_json(self):
        _json = dict()
        if self.item_max_days is not None:
            _json["itemMaxDays"] = self.item_max_days
        return _json

    @classmethod
    def from_json(cls, _json: dict):
        item_max_days = _json.get('itemMaxDays', None)
        if item_max_days:
            return cls(item_max_days=item_max_days)
        return None


@attr.s
class Dataset(entities.BaseEntity):
    """
    Dataset object
    """
    # dataset information
    id = attr.ib()
    url = attr.ib()
    name = attr.ib()
    annotated = attr.ib(repr=False)
    creator = attr.ib()
    projects = attr.ib(repr=False)
    items_count = attr.ib()
    metadata = attr.ib(repr=False)
    directoryTree = attr.ib(repr=False)
    expiration_options = attr.ib()
    index_driver = attr.ib()
    enable_sync_with_cloned = attr.ib(repr=False)

    # name change when to_json
    created_at = attr.ib()
    updated_at = attr.ib()
    updated_by = attr.ib()
    items_url = attr.ib(repr=False)
    readable_type = attr.ib(repr=False)
    access_level = attr.ib(repr=False)
    driver = attr.ib(repr=False)
    src_dataset = attr.ib(repr=False)
    _readonly = attr.ib(repr=False)
    annotations_count = attr.ib()

    # api
    _client_api = attr.ib(type=ApiClient, repr=False)

    # syncing status
    is_syncing = attr.ib(default=False, repr=False)

    # entities
    _project = attr.ib(default=None, repr=False)

    # repositories
    _datasets = attr.ib(repr=False, default=None)
    _repositories = attr.ib(repr=False)

    # defaults
    _ontology_ids = attr.ib(default=None, repr=False)
    _labels = attr.ib(default=None, repr=False)
    _directory_tree = attr.ib(default=None, repr=False)
    _recipe = attr.ib(default=None, repr=False)
    _ontology = attr.ib(default=None, repr=False)

    @property
    def itemsCount(self):
        return self.items_count

    @staticmethod
    def _protected_from_json(project: entities.Project,
                             _json: dict,
                             client_api: ApiClient,
                             datasets=None,
                             is_fetched=True):
        """
        Same as from_json but with try-except to catch if error

        :param project: dataset's project
        :param _json: _json response from host
        :param client_api: ApiClient entity
        :param datasets: Datasets repository
        :param is_fetched: is Entity fetched from Platform
        :return: Dataset object
        """
        try:
            dataset = Dataset.from_json(project=project,
                                        _json=_json,
                                        client_api=client_api,
                                        datasets=datasets,
                                        is_fetched=is_fetched)
            status = True
        except Exception:
            dataset = traceback.format_exc()
            status = False
        return status, dataset

    @classmethod
    def from_json(cls,
                  project: entities.Project,
                  _json: dict,
                  client_api: ApiClient,
                  datasets=None,
                  is_fetched=True):
        """
        Build a Dataset entity object from a json

        :param project: dataset's project
        :param dict _json: _json response from host
        :param client_api: ApiClient entity
        :param datasets: Datasets repository
        :param bool is_fetched: is Entity fetched from Platform
        :return: Dataset object
        :rtype: dtlpy.entities.dataset.Dataset
        """
        projects = _json.get('projects', None)
        if project is not None and projects is not None:
            if project.id not in projects:
                logger.warning('Dataset has been fetched from a project that is not in it projects list')
                project = None

        expiration_options = _json.get('expirationOptions', None)
        if expiration_options:
            expiration_options = ExpirationOptions.from_json(expiration_options)
        inst = cls(metadata=_json.get('metadata', None),
                   directoryTree=_json.get('directoryTree', None),
                   readable_type=_json.get('readableType', None),
                   access_level=_json.get('accessLevel', None),
                   created_at=_json.get('createdAt', None),
                   updated_at=_json.get('updatedAt', None),
                   updated_by=_json.get('updatedBy', None),
                   annotations_count=_json.get("annotationsCount", None),
                   items_count=_json.get('itemsCount', None),
                   annotated=_json.get('annotated', None),
                   readonly=_json.get('readonly', None),
                   projects=projects,
                   creator=_json.get('creator', None),
                   items_url=_json.get('items', None),
                   driver=_json.get('driver', None),
                   name=_json.get('name', None),
                   url=_json.get('url', None),
                   id=_json.get('id', None),
                   datasets=datasets,
                   client_api=client_api,
                   project=project,
                   expiration_options=expiration_options,
                   index_driver=_json.get('indexDriver', None),
                   enable_sync_with_cloned=_json.get('enableSyncWithCloned', None),
                   is_syncing=_json.get('isSyncing', False),
                   src_dataset=_json.get('srcDataset', None))
        inst.is_fetched = is_fetched
        return inst

    def to_json(self):
        """
        Returns platform _json format of object

        :return: platform json format of object
        :rtype: dict
        """
        _json = attr.asdict(self, filter=attr.filters.exclude(attr.fields(Dataset)._client_api,
                                                              attr.fields(Dataset)._project,
                                                              attr.fields(Dataset)._readonly,
                                                              attr.fields(Dataset)._datasets,
                                                              attr.fields(Dataset)._repositories,
                                                              attr.fields(Dataset)._ontology_ids,
                                                              attr.fields(Dataset)._labels,
                                                              attr.fields(Dataset)._recipe,
                                                              attr.fields(Dataset)._ontology,
                                                              attr.fields(Dataset)._directory_tree,
                                                              attr.fields(Dataset).access_level,
                                                              attr.fields(Dataset).readable_type,
                                                              attr.fields(Dataset).created_at,
                                                              attr.fields(Dataset).updated_at,
                                                              attr.fields(Dataset).updated_by,
                                                              attr.fields(Dataset).annotations_count,
                                                              attr.fields(Dataset).items_url,
                                                              attr.fields(Dataset).expiration_options,
                                                              attr.fields(Dataset).items_count,
                                                              attr.fields(Dataset).index_driver,
                                                              attr.fields(Dataset).enable_sync_with_cloned,
                                                              attr.fields(Dataset).is_syncing,
                                                              attr.fields(Dataset).src_dataset,
                                                              ))
        _json.update({'items': self.items_url})
        _json['readableType'] = self.readable_type
        _json['createdAt'] = self.created_at
        _json['updatedAt'] = self.updated_at
        _json['updatedBy'] = self.updated_by
        _json['annotationsCount'] = self.annotations_count
        _json['accessLevel'] = self.access_level
        _json['readonly'] = self._readonly
        _json['itemsCount'] = self.items_count
        _json['indexDriver'] = self.index_driver
        if self.expiration_options and self.expiration_options.to_json():
            _json['expirationOptions'] = self.expiration_options.to_json()
        if self.enable_sync_with_cloned is not None:
            _json['enableSyncWithCloned'] = self.enable_sync_with_cloned
        _json['isSyncing'] = self.is_syncing
        if self.src_dataset is not None:
            _json['srcDataset'] = self.src_dataset
        return _json

    @property
    def labels(self):
        if self._labels is None:
            self._labels = self._get_ontology().labels
        return self._labels

    @property
    def readonly(self):
        return self._readonly

    @property
    def platform_url(self):
        return self._client_api._get_resource_url("projects/{}/datasets/{}/items".format(self.project.id, self.id))

    @readonly.setter
    def readonly(self, state):
        import warnings
        warnings.warn("`readonly` flag on dataset is deprecated, doing nothing.", DeprecationWarning)

    @property
    def labels_flat_dict(self):
        return self._get_ontology().labels_flat_dict

    @property
    def instance_map(self) -> dict:
        return self._get_ontology().instance_map

    @instance_map.setter
    def instance_map(self, value: dict):
        """
        instance mapping for creating instance mask

        :param value: dictionary {label: map_id}
        """
        if not isinstance(value, dict):
            raise ValueError('input must be a dictionary of {label_name: instance_id}')
        self._get_ontology().instance_map = value

    @property
    def ontology_ids(self):
        if self._ontology_ids is None:
            self._ontology_ids = list()
            if self.metadata is not None and 'system' in self.metadata and 'recipes' in self.metadata['system']:
                recipe_ids = self.get_recipe_ids()
                for rec_id in recipe_ids:
                    recipe = self.recipes.get(recipe_id=rec_id)
                    self._ontology_ids += recipe.ontology_ids
        return self._ontology_ids

    @_repositories.default
    def set_repositories(self):
        reps = namedtuple('repositories',
                          field_names=['items', 'recipes', 'datasets', 'assignments', 'tasks', 'annotations',
                                       'ontologies', 'features', 'feature_sets', 'settings', 'schema', 'collections'])
        _project_id = None
        if self._project is None:
            datasets = repositories.Datasets(client_api=self._client_api, project=self._project)
            if self.projects is not None and len(self.projects) > 0:
                _project_id = self.projects[0]
        else:
            datasets = self._project.datasets
            _project_id = self._project.id
        return reps(
            items=repositories.Items(client_api=self._client_api, dataset=self, datasets=datasets),
            recipes=repositories.Recipes(client_api=self._client_api, dataset=self),
            assignments=repositories.Assignments(project=self._project, client_api=self._client_api, dataset=self),
            tasks=repositories.Tasks(client_api=self._client_api, project=self._project, dataset=self),
            annotations=repositories.Annotations(client_api=self._client_api, dataset=self),
            datasets=datasets,
            ontologies=repositories.Ontologies(client_api=self._client_api, dataset=self),
            features=repositories.Features(client_api=self._client_api, project=self._project, dataset=self),
            feature_sets=repositories.FeatureSets(client_api=self._client_api, project=self._project, project_id=_project_id, dataset=self),
            settings=repositories.Settings(client_api=self._client_api, dataset=self),
            schema=repositories.Schema(client_api=self._client_api, dataset=self),
            collections=repositories.Collections(client_api=self._client_api, dataset=self)
        )

    @property
    def settings(self):
        assert isinstance(self._repositories.settings, repositories.Settings)
        return self._repositories.settings

    @property
    def items(self):
        assert isinstance(self._repositories.items, repositories.Items)
        return self._repositories.items

    @property
    def ontologies(self):
        assert isinstance(self._repositories.ontologies, repositories.Ontologies)
        return self._repositories.ontologies

    @property
    def recipes(self):
        assert isinstance(self._repositories.recipes, repositories.Recipes)
        return self._repositories.recipes

    @property
    def datasets(self):
        assert isinstance(self._repositories.datasets, repositories.Datasets)
        return self._repositories.datasets

    @property
    def assignments(self):
        assert isinstance(self._repositories.assignments, repositories.Assignments)
        return self._repositories.assignments

    @property
    def tasks(self):
        assert isinstance(self._repositories.tasks, repositories.Tasks)
        return self._repositories.tasks

    @property
    def annotations(self):
        assert isinstance(self._repositories.annotations, repositories.Annotations)
        return self._repositories.annotations

    @property
    def features(self):
        assert isinstance(self._repositories.features, repositories.Features)
        return self._repositories.features

    @property
    def feature_sets(self):
        assert isinstance(self._repositories.feature_sets, repositories.FeatureSets)
        return self._repositories.feature_sets

    @property
    def collections(self):
        assert isinstance(self._repositories.collections, repositories.Collections)
        return self._repositories.collections

    @property
    def schema(self):
        assert isinstance(self._repositories.schema, repositories.Schema)
        return self._repositories.schema

    @property
    def project(self):
        if self._project is None:
            # get from cache
            project = self._client_api.state_io.get('project')
            if project is not None:
                # build entity from json
                p = entities.Project.from_json(_json=project, client_api=self._client_api)
                # check if dataset belongs to project
                if p.id in self.projects:
                    self._project = p
        if self._project is None:
            self._project = repositories.Projects(client_api=self._client_api).get(project_id=self.projects[0],
                                                                                   fetch=None)
        assert isinstance(self._project, entities.Project)
        return self._project

    @project.setter
    def project(self, project):
        if not isinstance(project, entities.Project):
            raise ValueError('Must input a valid Project entity')
        self._project = project

    @property
    def directory_tree(self):
        if self._directory_tree is None:
            self._directory_tree = self.project.datasets.directory_tree(dataset_id=self.id)
        assert isinstance(self._directory_tree, entities.DirectoryTree)
        return self._directory_tree

    def __copy__(self):
        return Dataset.from_json(_json=self.to_json(),
                                 project=self._project,
                                 client_api=self._client_api,
                                 is_fetched=self.is_fetched,
                                 datasets=self.datasets)

    def __get_local_path__(self):
        if self._project is not None:
            local_path = os.path.join(services.service_defaults.DATALOOP_PATH,
                                      'projects',
                                      self.project.name,
                                      'datasets',
                                      self.name)
        else:
            local_path = os.path.join(services.service_defaults.DATALOOP_PATH,
                                      'datasets',
                                      '%s_%s' % (self.name, self.id))
        return local_path

    def _get_recipe(self):
        recipes = self.recipes.list()
        if len(recipes) > 0:
            return recipes[0]
        else:
            raise exceptions.PlatformException('404', 'Dataset {} has no recipe'.format(self.name))

    def _get_ontology(self):
        if self._ontology is None:
            ontologies = self._get_recipe().ontologies.list()
            if len(ontologies) > 0:
                self._ontology = ontologies[0]
            else:
                raise exceptions.PlatformException('404', 'Dataset {} has no ontology'.format(self.name))
        return self._ontology

    @staticmethod
    def serialize_labels(labels_dict):
        """
        Convert hex color format to rgb

        :param dict labels_dict: dict of labels
        :return: dict of converted labels
        """
        dataset_labels_dict = dict()
        for label, color in labels_dict.items():
            dataset_labels_dict[label] = '#%02x%02x%02x' % color
        return dataset_labels_dict

    def get_recipe_ids(self):
        """
        Get dataset recipe Ids

        :return: list of recipe ids
        :rtype: list
        """
        return self.metadata['system']['recipes']

    def switch_recipe(self, recipe_id=None, recipe=None):
        """
        Switch the recipe that linked to the dataset with the given one

        :param str recipe_id: recipe id
        :param dtlpy.entities.recipe.Recipe recipe: recipe entity

        **Example**:

        .. code-block:: python

            dataset.switch_recipe(recipe_id='recipe_id')
        """
        if recipe is None and recipe_id is None:
            raise exceptions.PlatformException('400', 'Must provide recipe or recipe_id')
        if recipe_id is None:
            if not isinstance(recipe, entities.Recipe):
                raise exceptions.PlatformException('400', 'Recipe must me entities.Recipe type')
            else:
                recipe_id = recipe.id

        # add recipe id to dataset metadata
        if 'system' not in self.metadata:
            self.metadata['system'] = dict()
        if 'recipes' not in self.metadata['system']:
            self.metadata['system']['recipes'] = list()
        self.metadata['system']['recipes'] = [recipe_id]
        self.update(system_metadata=True)

    def delete(self, sure=False, really=False):
        """
        Delete a dataset forever!

        **Prerequisites**: You must be an *owner* or *developer* to use this method.

        :param bool sure: are you sure you want to delete?
        :param bool really: really really?
        :return: True is success
        :rtype: bool

        **Example**:

        .. code-block:: python

            is_deleted = dataset.delete(sure=True, really=True)
        """
        return self.datasets.delete(dataset_id=self.id,
                                    sure=sure,
                                    really=really)

    def update(self, system_metadata=False):
        """
        Update dataset field

        **Prerequisites**: You must be an *owner* or *developer* to use this method.

        :param bool system_metadata: bool - True, if you want to change metadata system
        :return: Dataset object
        :rtype: dtlpy.entities.dataset.Dataset

        **Example**:

        .. code-block:: python

            dataset = dataset.update()
        """
        return self.datasets.update(dataset=self,
                                    system_metadata=system_metadata)

    def unlock(self):
        """
        Unlock dataset

        **Prerequisites**: You must be an *owner* or *developer* to use this method.

        :return: Dataset object
        :rtype: dtlpy.entities.dataset.Dataset

        **Example**:

        .. code-block:: python

            dataset = dataset.unlock()
        """
        return self.datasets.unlock(dataset=self)

    def set_readonly(self, state: bool):
        """
        Set dataset readonly mode

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param bool state: state

        **Example**:

        .. code-block:: python

            dataset.set_readonly(state=True)
        """
        import warnings
        warnings.warn("`readonly` flag on dataset is deprecated, doing nothing.", DeprecationWarning)

    def clone(self,
              clone_name=None,
              filters=None,
              with_items_annotations=True,
              with_metadata=True,
              with_task_annotations_status=True,
              dst_dataset_id=None,
              target_directory=None,
              ):
        """
        Clone dataset

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param str clone_name: new dataset name
        :param dtlpy.entities.filters.Filters filters: Filters entity or a query dict
        :param bool with_items_annotations: clone all item's annotations
        :param bool with_metadata: clone metadata
        :param bool with_task_annotations_status: clone task annotations status
        :param str dst_dataset_id: destination dataset id
        :param str target_directory: target directory
        :return: dataset object
        :rtype: dtlpy.entities.dataset.Dataset

        **Example**:

        .. code-block:: python

            dataset = dataset.clone(dataset_id='dataset_id',
                          clone_name='dataset_clone_name',
                          with_metadata=True,
                          with_items_annotations=False,
                          with_task_annotations_status=False)
        """
        return self.datasets.clone(dataset_id=self.id,
                                   filters=filters,
                                   clone_name=clone_name,
                                   with_metadata=with_metadata,
                                   with_items_annotations=with_items_annotations,
                                   with_task_annotations_status=with_task_annotations_status,
                                   dst_dataset_id=dst_dataset_id,
                                   target_directory=target_directory)

    def sync(self, wait=True):
        """
        Sync dataset with external storage

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param bool wait: wait for the command to finish
        :return: True if success
        :rtype: bool

        **Example**:

        .. code-block:: python

            success = dataset.sync()
        """
        return self.datasets.sync(dataset_id=self.id, wait=wait)

    def download_annotations(self,
                             local_path=None,
                             filters=None,
                             annotation_options: ViewAnnotationOptions = None,
                             annotation_filters=None,
                             overwrite=False,
                             thickness=1,
                             with_text=False,
                             remote_path=None,
                             include_annotations_in_output=True,
                             export_png_files=False,
                             filter_output_annotations=False,
                             alpha=1,
                             export_version=ExportVersion.V1,
                             dataset_lock=False,
                             lock_timeout_sec=None,
                             export_summary=False,
                            ):
        """
        Download dataset by filters.
        Filtering the dataset for items and save them local
        Optional - also download annotation, mask, instance and image mask of the item

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param str local_path: local folder or filename to save to.
        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :param list(dtlpy.entities.annotation.ViewAnnotationOptions) annotation_options: download annotations options: list(dl.ViewAnnotationOptions)
        :param dtlpy.entities.filters.Filters annotation_filters: Filters entity to filter annotations for download
        :param bool overwrite: optional - default = False
        :param bool dataset_lock: optional - default = False
        :param bool export_summary: optional - default = False
        :param int lock_timeout_sec: optional
        :param int thickness: optional - line thickness, if -1 annotation will be filled, default =1
        :param bool with_text: optional - add text to annotations, default = False
        :param str remote_path: DEPRECATED and ignored
        :param bool include_annotations_in_output: default - False , if export should contain annotations
        :param bool export_png_files: default - if True, semantic annotations should be exported as png files
        :param bool filter_output_annotations: default - False, given an export by filter - determine if to filter out annotations
        :param float alpha: opacity value [0 1], default 1
        :param str export_version:  exported items will have original extension in filename, `V1` - no original extension in filenames
        :return: local_path of the directory where all the downloaded item
        :rtype: str

        **Example**:

        .. code-block:: python

            local_path = dataset.download_annotations(dataset='dataset_entity',
                                         local_path='local_path',
                                         annotation_options=[dl.ViewAnnotationOptions.JSON, dl.ViewAnnotationOptions.MASK],
                                         overwrite=False,
                                         thickness=1,
                                         with_text=False,
                                         alpha=1,
                                         dataset_lock=False,
                                         lock_timeout_sec=300,
                                         export_summary=False
                                        )
        """

        return self.datasets.download_annotations(
            dataset=self,
            local_path=local_path,
            overwrite=overwrite,
            filters=filters,
            annotation_options=annotation_options,
            annotation_filters=annotation_filters,
            thickness=thickness,
            with_text=with_text,
            remote_path=remote_path,
            include_annotations_in_output=include_annotations_in_output,
            export_png_files=export_png_files,
            filter_output_annotations=filter_output_annotations,
            alpha=alpha,
            export_version=export_version,
            dataset_lock=dataset_lock,
            lock_timeout_sec=lock_timeout_sec,
            export_summary=export_summary        
        )

    def export(self,
               local_path=None,
               filters=None,
               annotation_filters=None,
               feature_vector_filters=None,
               include_feature_vectors: bool = False,
               include_annotations: bool = False,
               export_type: ExportType = ExportType.JSON,
               timeout: int = 0,
               dataset_lock: bool = False,
               lock_timeout_sec: int = None,
               export_summary: bool = False,
               output_export_type: OutputExportType = None):
        """
        Export dataset items and annotations.

        **Prerequisites**: You must be an *owner* or *developer* to use this method.

        You must provide at least ONE of the following params: dataset, dataset_name, dataset_id.

        :param str local_path: The local path to save the exported dataset
        :param Union[dict, dtlpy.entities.filters.Filters] filters: Filters entity or a query dictionary
        :param dtlpy.entities.filters.Filters annotation_filters: Filters entity
        :param dtlpy.entities.filters.Filters feature_vector_filters: Filters entity
        :param bool include_feature_vectors: Include item feature vectors in the export
        :param bool include_annotations: Include item annotations in the export
        :param bool dataset_lock: Make dataset readonly during the export
        :param bool export_summary: Download dataset export summary
        :param int lock_timeout_sec: Timeout for locking the dataset during export in seconds
        :param entities.ExportType export_type: Type of export ('json' or 'zip')
        :param entities.OutputExportType output_export_type: Output format ('json', 'zip', or 'folders'). If None, defaults to 'json'
        :param int timeout: Maximum time in seconds to wait for the export to complete
        :return: Exported item
        :rtype: dtlpy.entities.item.Item

        **Example**:

        .. code-block:: python

            export_item = dataset.export(filters=filters,
                                         include_feature_vectors=True,
                                         include_annotations=True,
                                         export_type=dl.ExportType.JSON,
                                         output_export_type=dl.OutputExportType.JSON)
        """

        return self.datasets.export(dataset=self,
                                    local_path=local_path,
                                    filters=filters,
                                    annotation_filters=annotation_filters,
                                    feature_vector_filters=feature_vector_filters,
                                    include_feature_vectors=include_feature_vectors,
                                    include_annotations=include_annotations,
                                    export_type=export_type,
                                    timeout=timeout,
                                    dataset_lock=dataset_lock,
                                    lock_timeout_sec=lock_timeout_sec,
                                    export_summary=export_summary,
                                    output_export_type=output_export_type)

    def upload_annotations(self,
                           local_path,
                           filters=None,
                           clean=False,
                           remote_root_path='/',
                           export_version=ExportVersion.V1
                           ):
        """
        Upload annotations to dataset.

        **Prerequisites**: You must have a dataset with items that are related to the annotations. The relationship between the dataset and annotations is shown in the name. You must be in the role of an *owner* or *developer*.

        :param str local_path: str - local folder where the annotations files is.
        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :param bool clean: bool - if True it remove the old annotations
        :param str remote_root_path: str - the remote root path to match remote and local items
        :param str export_version:  `V2` - exported items will have original extension in filename, `V1` - no original extension in filenames

        For example, if the item filepath is a/b/item and remote_root_path is /a the start folder will be b instead of a

        **Example**:

        .. code-block:: python

            dataset.upload_annotations(dataset='dataset_entity',
                                     local_path='local_path',
                                     clean=False,
                                     export_version=dl.ExportVersion.V1
                                     )
        """

        return self.datasets.upload_annotations(
            dataset=self,
            local_path=local_path,
            filters=filters,
            clean=clean,
            remote_root_path=remote_root_path,
            export_version=export_version
        )

    def checkout(self):
        """
        Checkout the dataset

        """
        self.datasets.checkout(dataset=self)

    def open_in_web(self):
        """
        Open the dataset in web platform

        """
        self._client_api._open_in_web(url=self.platform_url)

    def add_label(self, label_name, color=None, children=None, attributes=None, display_label=None, label=None,
                  recipe_id=None, ontology_id=None, icon_path=None):
        """
        Add single label to dataset

        **Prerequisites**: You must have a dataset with items that are related to the annotations. The relationship between the dataset and annotations is shown in the name. You must be in the role of an *owner* or *developer*.

        :param str label_name: str - label name
        :param tuple color: RGB color of the annotation, e.g (255,0,0) or '#ff0000' for red
        :param children: children (sub labels). list of sub labels of this current label, each value is either dict or dl.Label
        :param list attributes: add attributes to the labels
        :param str display_label: name that display label
        :param dtlpy.entities.label.Label label: label object
        :param str recipe_id: optional recipe id
        :param str ontology_id: optional ontology id
        :param str icon_path: path to image to be display on label
        :return: label entity
        :rtype: dtlpy.entities.label.Label

        **Example**:

        .. code-block:: python

            dataset.add_label(label_name='person', color=(34, 6, 231), attributes=['big', 'small'])
        """
        # get recipe
        if recipe_id is None:
            recipe_id = self.get_recipe_ids()[0]
        recipe = self.recipes.get(recipe_id=recipe_id)

        # get ontology
        if ontology_id is None:
            ontology_id = recipe.ontology_ids[0]
        ontology = recipe.ontologies.get(ontology_id=ontology_id)
        # ontology._dataset = self

        # add label
        added_label = ontology.add_label(label_name=label_name,
                                         color=color,
                                         children=children,
                                         attributes=attributes,
                                         display_label=display_label,
                                         label=label,
                                         update_ontology=True,
                                         icon_path=icon_path)

        return added_label

    def add_labels(self, label_list, ontology_id=None, recipe_id=None):
        """
        Add labels to dataset

        **Prerequisites**: You must have a dataset with items that are related to the annotations. The relationship between the dataset and annotations is shown in the name. You must be in the role of an *owner* or *developer*.

        :param list label_list: a list of labels to add to the dataset's ontology. each value should be a dict, dl.Label or a string
        :param str ontology_id: optional ontology id
        :param str recipe_id: optional recipe id
        :return: label entities

        **Example**:

        .. code-block:: python

            dataset.add_labels(label_list=label_list)
        """
        # get recipe
        if recipe_id is None:
            recipe_id = self.get_recipe_ids()[0]
        recipe = self.recipes.get(recipe_id=recipe_id)

        # get ontology
        if ontology_id is None:
            ontology_id = recipe.ontology_ids[0]
        ontology = recipe.ontologies.get(ontology_id=ontology_id)

        # add labels to ontology
        added_labels = ontology.add_labels(label_list=label_list, update_ontology=True)

        return added_labels

    def update_label(self, label_name, color=None, children=None, attributes=None, display_label=None, label=None,
                     recipe_id=None, ontology_id=None, upsert=False, icon_path=None):
        """
        Add single label to dataset

        **Prerequisites**: You must have a dataset with items that are related to the annotations. The relationship between the dataset and annotations is shown in the name. You must be in the role of an *owner* or *developer*.

        :param str label_name: str - label name
        :param tuple color: color
        :param children: children (sub labels)
        :param list attributes: add attributes to the labels
        :param str display_label: name that display label
        :param dtlpy.entities.label.Label label: label
        :param str recipe_id: optional recipe id
        :param str ontology_id: optional ontology id
        :param str icon_path: path to image to be display on label

        :return: label entity
        :rtype: dtlpy.entities.label.Label

        **Example**:

        .. code-block:: python

            dataset.update_label(label_name='person', color=(34, 6, 231), attributes=['big', 'small'])
        """
        # get recipe

        if recipe_id is None:
            recipe_id = self.get_recipe_ids()[0]
        recipe = self.recipes.get(recipe_id=recipe_id)

        # get ontology
        if ontology_id is None:
            ontology_id = recipe.ontology_ids[0]
        ontology = recipe.ontologies.get(ontology_id=ontology_id)

        # add label
        added_label = ontology.update_label(label_name=label_name,
                                            color=color,
                                            children=children,
                                            attributes=attributes,
                                            display_label=display_label,
                                            label=label,
                                            update_ontology=True,
                                            upsert=upsert,
                                            icon_path=icon_path)

        return added_label

    def update_labels(self, label_list, ontology_id=None, recipe_id=None, upsert=False):
        """
        Add labels to dataset

        **Prerequisites**: You must have a dataset with items that are related to the annotations. The relationship between the dataset and annotations is shown in the name. You must be in the role of an *owner* or *developer*.

        :param list label_list: label list
        :param str ontology_id: optional ontology id
        :param str recipe_id: optional recipe id
        :param bool upsert: if True will add in case it does not existing

        :return: label entities
        :rtype: dtlpy.entities.label.Label

        **Example**:

        .. code-block:: python

            dataset.update_labels(label_list=label_list)
        """
        # get recipe
        if recipe_id is None:
            recipe_id = self.get_recipe_ids()[0]
        recipe = self.recipes.get(recipe_id=recipe_id)

        # get ontology
        if ontology_id is None:
            ontology_id = recipe.ontology_ids[0]
        ontology = recipe.ontologies.get(ontology_id=ontology_id)

        # add labels to ontology
        added_labels = ontology.update_labels(label_list=label_list, update_ontology=True, upsert=upsert)

        return added_labels

    def download(
            self,
            filters=None,
            local_path=None,
            file_types=None,
            annotation_options: ViewAnnotationOptions = None,
            annotation_filters=None,
            overwrite=False,
            to_items_folder=True,
            thickness=1,
            with_text=False,
            without_relative_path=None,
            alpha=1,
            export_version=ExportVersion.V1,
            dataset_lock=False,
            lock_timeout_sec=None,
            export_summary=False,
            raise_on_error=False
    ):
        """
        Download dataset by filters.
        Filtering the dataset for items and save them local
        Optional - also download annotation, mask, instance and image mask of the item

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :param str local_path: local folder or filename to save to.
        :param list file_types: a list of file type to download. e.g ['video/webm', 'video/mp4', 'image/jpeg', 'image/png']
        :param list annotation_options: type of download annotations: list(dl.ViewAnnotationOptions)
        :param dtlpy.entities.filters.Filters annotation_filters: Filters entity to filter annotations for download
        :param bool overwrite: optional - default = False to overwrite the existing files
        :param bool dataset_lock: optional - default = False to make dataset readonly during the download
        :param bool export_summary: optional - default = False to get the symmary of the export
        :param int lock_timeout_sec: optional - Set lock timeout for the export
        :param bool to_items_folder: Create 'items' folder and download items to it
        :param int thickness: optional - line thickness, if -1 annotation will be filled, default =1
        :param bool with_text: optional - add text to annotations, default = False
        :param bool without_relative_path: bool - download items without the relative path from platform
        :param float alpha: opacity value [0 1], default 1
        :param str export_version:  `V2` - exported items will have original extension in filename, `V1` - no original extension in filenames
        :param bool raise_on_error: raise an exception if an error occurs
        :return: `List` of local_path per each downloaded item

        **Example**:

        .. code-block:: python

            dataset.download(local_path='local_path',
                             annotation_options=[dl.ViewAnnotationOptions.JSON, dl.ViewAnnotationOptions.MASK],
                             overwrite=False,
                             thickness=1,
                             with_text=False,
                             alpha=1,
                             dataset_lock=False,      
                             lock_timeout_sec=300,
                             export_summary=False                      
                             )
        """
        return self.items.download(filters=filters,
                                   local_path=local_path,
                                   file_types=file_types,
                                   annotation_options=annotation_options,
                                   annotation_filters=annotation_filters,
                                   overwrite=overwrite,
                                   to_items_folder=to_items_folder,
                                   thickness=thickness,
                                   with_text=with_text,
                                   without_relative_path=without_relative_path,
                                   alpha=alpha,
                                   export_version=export_version,
                                   dataset_lock=dataset_lock,
                                   lock_timeout_sec=lock_timeout_sec,
                                   export_summary=export_summary,
                                   raise_on_error=raise_on_error
                                   )

    def download_folder(
            self,
            folder_path,
            filters=None,
            local_path=None,
            file_types=None,
            annotation_options: ViewAnnotationOptions = None,
            annotation_filters=None,
            overwrite=False,
            to_items_folder=True,
            thickness=1,
            with_text=False,
            without_relative_path=None,
            alpha=1,
            export_version=ExportVersion.V1,
            dataset_lock=False,
            lock_timeout_sec=None,
            export_summary=False,
            raise_on_error=False
    ):
        """
        Download dataset folder.
        Optional - also download annotation, mask, instance and image mask of the item

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param str folder_path: the path of the folder that want to download
        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :param str local_path: local folder or filename to save to.
        :param list file_types: a list of file type to download. e.g ['video/webm', 'video/mp4', 'image/jpeg', 'image/png']
        :param list annotation_options: type of download annotations: list(dl.ViewAnnotationOptions)
        :param dtlpy.entities.filters.Filters annotation_filters: Filters entity to filter annotations for download
        :param bool overwrite: optional - default = False to overwrite the existing files
        :param bool dataset_lock: optional - default = False to make the dataset readonly during the download
        :param bool export_summary: optional - default = False to get the symmary of the export
        :param bool lock_timeout_sec: optional - Set lock timeout for the export
        :param bool to_items_folder: Create 'items' folder and download items to it
        :param int thickness: optional - line thickness, if -1 annotation will be filled, default =1
        :param bool with_text: optional - add text to annotations, default = False
        :param bool without_relative_path: bool - download items without the relative path from platform
        :param float alpha: opacity value [0 1], default 1
        :param str export_version:  `V2` - exported items will have original extension in filename, `V1` - no original extension in filenames
        :param bool raise_on_error: raise an exception if an error occurs
        :return: `List` of local_path per each downloaded item

        **Example**:

        .. code-block:: python

            dataset.download_folder(folder_path='folder_path'
                             local_path='local_path',
                             annotation_options=[dl.ViewAnnotationOptions.JSON, dl.ViewAnnotationOptions.MASK],
                             overwrite=False,
                             thickness=1,
                             with_text=False,
                             alpha=1,
                             save_locally=True,
                             dataset_lock=False
                             lock_timeout_sec=300,
                             export_summary=False
                             )
        """
        filters = self.datasets._bulid_folder_filter(folder_path=folder_path, filters=filters)
        return self.items.download(filters=filters,
                                   local_path=local_path,
                                   file_types=file_types,
                                   annotation_options=annotation_options,
                                   annotation_filters=annotation_filters,
                                   overwrite=overwrite,
                                   to_items_folder=to_items_folder,
                                   thickness=thickness,
                                   with_text=with_text,
                                   without_relative_path=without_relative_path,
                                   alpha=alpha,
                                   export_version=export_version,
                                   dataset_lock=dataset_lock,
                                   lock_timeout_sec=lock_timeout_sec,
                                    export_summary=export_summary,
                                    raise_on_error=raise_on_error
                                   )

    def delete_labels(self, label_names):
        """
        Delete labels from dataset's ontologies

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param label_names: label object/ label name / list of label objects / list of label names

        **Example**:

        .. code-block:: python

            dataset.delete_labels(label_names=['myLabel1', 'Mylabel2'])
        """
        for recipe in self.recipes.list():
            for ontology in recipe.ontologies.list():
                ontology.delete_labels(label_names=label_names)
        self._labels = None

    def update_attributes(self,
                          title: str,
                          key: str,
                          attribute_type,
                          recipe_id: str = None,
                          ontology_id: str = None,
                          scope: list = None,
                          optional: bool = None,
                          values: list = None,
                          attribute_range=None):
        """
        ADD a new attribute or update if exist

        :param str ontology_id: ontology_id
        :param str title: attribute title
        :param str key: the key of the attribute must br unique
        :param AttributesTypes attribute_type: dl.AttributesTypes your attribute type
        :param list scope: list of the labels or * for all labels
        :param bool optional: optional attribute
        :param list values: list of the attribute values ( for checkbox and radio button)
        :param dict or AttributesRange attribute_range: dl.AttributesRange object
        :return: true in success
        :rtype: bool

        **Example**:

        .. code-block:: python

            dataset.update_attributes(ontology_id='ontology_id',
                                      key='1',
                                      title='checkbox',
                                      attribute_type=dl.AttributesTypes.CHECKBOX,
                                      values=[1,2,3])
        """
        # get recipe
        if recipe_id is None:
            recipe_id = self.get_recipe_ids()[0]
        recipe = self.recipes.get(recipe_id=recipe_id)

        # get ontology
        if ontology_id is None:
            ontology_id = recipe.ontology_ids[0]
        ontology = recipe.ontologies.get(ontology_id=ontology_id)

        # add attribute to ontology
        attribute = ontology.update_attributes(
            title=title,
            key=key,
            attribute_type=attribute_type,
            scope=scope,
            optional=optional,
            values=values,
            attribute_range=attribute_range)

        return attribute

    def delete_attributes(self, keys: list,
                          recipe_id: str = None,
                          ontology_id: str = None):
        """
        Delete a bulk of attributes

        :param str recipe_id: recipe id
        :param str ontology_id: ontology id
        :param list keys: Keys of attributes to delete
        :return: True if success
        :rtype: bool
        """

        # get recipe
        if recipe_id is None:
            recipe_id = self.get_recipe_ids()[0]
        recipe = self.recipes.get(recipe_id=recipe_id)

        # get ontology
        if ontology_id is None:
            ontology_id = recipe.ontology_ids[0]
        ontology = recipe.ontologies.get(ontology_id=ontology_id)
        return ontology.delete_attributes(ontology_id=ontology.id, keys=keys)

    def split_ml_subsets(self,
                         items_query = None,
                         percentages: dict = None ):
        """
        Split dataset items into ML subsets.

        :param dl.Filters items_query: Filters object to select items.
        :param dict percentages: {'train': x, 'validation': y, 'test': z}.
        :return: True if the split operation was successful.
        :rtype: bool
        """
        return self.datasets.split_ml_subsets(dataset_id=self.id,
                                              items_query=items_query,
                                              ml_split_list=percentages)

    def assign_subset_to_items(self, subset: str, items_query=None) -> bool:
        """
        Assign a specific ML subset (train/validation/test) to items defined by the given filters.
        This will set the chosen subset to True and the others to None.

        :param dl.Filters items_query: Filters to select items
        :param str subset: 'train', 'validation', or 'test'
        :return: True if successful
        :rtype: bool
        """
   
        return self.datasets.bulk_update_ml_subset(dataset_id=self.id,
                                                   items_query=items_query,
                                                   subset=subset)

    def remove_subset_from_items(self, items_query= None,) -> bool:
        """
        Remove any ML subset assignment from items defined by the given filters.
        This sets train, validation, and test tags to None.

        :param dl.Filters items_query: Filters to select items
        :return: True if successful
        :rtype: bool
        """
        return self.datasets.bulk_update_ml_subset(dataset_id=self.id,
                                                   items_query=items_query,
                                                   subset=None,
                                                   deleteTag=True)

    def get_items_missing_ml_subset(self, filters = None) -> list:
        """
        Get the list of item IDs that are missing ML subset assignment.
        An item is considered missing ML subset if train, validation, and test tags are not True (all None).

        :param dl.Filters filters: optional filters to narrow down items. If None, will use a default filter for files.
        :return: list of item IDs
        :rtype: list
        """
        if filters is None:
            filters = entities.Filters()
        filters.add(field='metadata.system.tags.train', values=None)
        filters.add(field='metadata.system.tags.validation', values=None)
        filters.add(field='metadata.system.tags.test', values=None)
        missing_ids = []
        pages = self.items.list(filters=filters)
        for page in pages:
            for item in page:
                # item that pass filters means no subsets assigned
                missing_ids.append(item.id)
        return missing_ids



import logging
from .. import entities, repositories, exceptions, miscellaneous, _api_reference
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')


class Compositions:
    def __init__(self, client_api: ApiClient, project: entities.Project = None):
        self._client_api = client_api
        self._project = project

    ############
    # entities #
    ############
    @property
    def project(self) -> entities.Project:
        if self._project is None:
            try:
                self._project = repositories.Projects(client_api=self._client_api).get()
            except exceptions.NotFound:
                raise exceptions.PlatformException(
                    error='2001',
                    message='Missing "project". need to set a Project entity or use project.pipelines repository')
        return self._project

    @project.setter
    def project(self, project: entities.Project):
        if not isinstance(project, entities.Project):
            raise ValueError('Must input a valid Project entity')
        self._project = project

    def get(self,
            composition_id=None,
            fetch=None
            ) -> entities.Pipeline:

        if fetch is None:
            fetch = self._client_api.fetch_entities

        if composition_id is None:
            raise exceptions.PlatformException(
                error='400',
                message='Must provide an identifier in inputs')

        elif fetch:
            if composition_id is not None:
                success, response = self._client_api.gen_request(
                    req_type="get",
                    path="/compositions/{}".format(composition_id))
                if not success:
                    raise exceptions.PlatformException(response)
                composition = response.json()
            else:
                raise exceptions.PlatformException(
                    error='400',
                    message='No checked-out pipeline was found, must checkout or provide an identifier in inputs')
        else:
            composition = {'id': composition_id}

        return composition



import datetime
import threading
import logging.handlers
import os

from .service_defaults import DATALOOP_PATH

logger = logging.getLogger(name='dtlpy')


class DataloopLogger(logging.handlers.BaseRotatingHandler):
    """
        Based on logging.handlers.RotatingFileHandler
        Create a new log file after reached maxBytes
        Delete logs older than a threshold default is week)
    """

    def __init__(self, filename, mode='a', maxBytes=0, encoding='utf-8', delay=False):
        if maxBytes > 0:
            mode = 'a'
        super().__init__(filename=filename, mode=mode, encoding=encoding, delay=delay)
        self.maxBytes = maxBytes
        DataloopLogger.clean_dataloop_cache()

    @staticmethod
    def clean_dataloop_cache(cache_path=DATALOOP_PATH, max_param=None):
        try:
            async_clean = True
            dir_list = [os.path.join(cache_path, d) for d in os.listdir(cache_path)
                        if os.path.isdir(os.path.join(cache_path, d))]
            for path in dir_list:
                if 'cache' not in path:
                    if async_clean:
                        worker = threading.Thread(target=DataloopLogger.clean_dataloop_cache_thread,
                                                  kwargs={'path': path,
                                                          'max_param': max_param})
                        worker.daemon = True
                        worker.start()
                    else:
                        DataloopLogger.clean_dataloop_cache_thread(path=path, max_param=max_param)
        except Exception as err:
            logger.exception(err)

    @staticmethod
    def get_clean_parameter_per(path):
        # (60 * 60 * 24 * 7):  # sec * min * hour * days - delete if older than a week
        # 1e6 100MB
        path_param = [{'type': 'datasets', 'max_time': 60 * 60 * 24 * 30},
                      {'type': 'items', 'max_time': 60 * 60 * 24 * 30},
                      {'type': 'logs', 'max_time': 60 * 60 * 24 * 7, 'max_size': 200 * 1e6},
                      {'type': 'projects', 'max_time': 60 * 60 * 24 * 30}]
        for param in path_param:
            if param['type'] in path:
                return param
        return {'type': 'default', 'max_time': 60 * 60 * 24 * 30}

    @staticmethod
    def clean_dataloop_cache_thread(path, total_cache_size=0, max_param=None):
        try:
            is_root = False
            if max_param is None:
                max_param = DataloopLogger.get_clean_parameter_per(path)
                is_root = True

            now = datetime.datetime.timestamp(datetime.datetime.now())
            files = [os.path.join(path, f) for f in os.listdir(path)]
            files.sort(key=lambda x: -os.path.getmtime(x))  # newer first
            for filepath in files:
                if os.path.isdir(filepath):
                    total_cache_size = DataloopLogger. \
                        clean_dataloop_cache_thread(filepath, total_cache_size=total_cache_size, max_param=max_param)
                    # Remove the dir if empty
                    if len(os.listdir(filepath)) == 0:
                        os.rmdir(filepath)
                    continue
                if 'max_time' in max_param:
                    file_time = os.path.getmtime(filepath)
                    if (now - file_time) > max_param['max_time']:
                        try:
                            os.remove(filepath)
                        except Exception as e:
                            logger.warning("Old log file can not be removed: {}".format(e))
                        continue
                if 'max_size' in max_param:
                    file_size = os.path.getsize(filepath)
                    if (total_cache_size + file_size) > max_param['max_size']:
                        try:
                            os.remove(filepath)
                        except Exception as e:
                            logger.warning("Old log file can not be removed: {}".format(e))
                        continue
                    total_cache_size += file_size
            if is_root:
                logger.debug("clean_dataloop_cache_thread for {} directory has been ended".format(path))
            return total_cache_size
        except Exception as err:
            logger.exception(err)

    @staticmethod
    def get_log_path():
        log_path = os.path.join(DATALOOP_PATH, 'logs')
        if not os.path.isdir(log_path):
            os.makedirs(log_path, exist_ok=True)
        return log_path

    @staticmethod
    def get_log_filepath():
        log_path = DataloopLogger.get_log_path()
        log_filepath = os.path.join(log_path, '{}.log'.format(datetime.datetime.now(datetime.timezone.utc).strftime('%Y-%m-%d_%H-%M-%S')))
        return log_filepath

    def doRollover(self):
        """
        Do a rollover, as described in __init__().
        """
        if self.stream:
            self.stream.close()
            self.stream = None
        # clean older logs (week old)
        DataloopLogger.clean_dataloop_cache()
        # create new log
        self.baseFilename = DataloopLogger.get_log_filepath()
        if not self.delay:
            self.stream = self._open()

    def shouldRollover(self, record):
        """
        Determine if rollover should occur.

        Basically, see if the supplied record would cause the file to exceed
        the size limit we have.
        """
        if self.stream is None:  # delay was set...
            self.stream = self._open()
        if self.maxBytes > 0:  # are we rolling over?
            msg = "%s\n" % self.format(record)
            self.stream.seek(0, 2)  # due to non-posix-compliant Windows feature
            if self.stream.tell() + len(msg) >= self.maxBytes:
                return 1
        return 0


class DtlpyFilter(logging.Filter):
    def __init__(self, package_path):
        super(DtlpyFilter, self).__init__(name='dtlpy')
        self._package_path = package_path

    def filter(self, record):
        pathname = record.pathname
        try:
            relativepath = os.path.splitext(os.path.relpath(pathname, self._package_path))[0]
            relativepath = relativepath.replace(os.sep, '.')
        except Exception:
            relativepath = ''
        record.relativepath = relativepath
        return True



import logging

from .. import exceptions, entities, miscellaneous
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')


class ResourceExecutions:
    """
    Resource Executions Repository

    The ResourceExecutions class allows the users to manage executions (executions of Resource) and their properties.
    """

    def __init__(self,
                 client_api: ApiClient,
                 project: entities.Project = None,
                 resource=None):
        self._client_api = client_api
        self._project = project
        self._resource = resource

    ############
    # entities #
    ############
    @property
    def resource(self):
        if self._resource is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Missing "resource". need to set a entity or use resource.executions repository')
        assert hasattr(entities, self._resource.__class__.__name__)
        return self._resource

    @resource.setter
    def resource(self, resource):
        if not hasattr(entities, self._resource.__class__.__name__):
            raise ValueError('Must input a valid entity')
        self._resource = resource

    @property
    def project(self) -> entities.Project:
        if self._project is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Missing "project". need to set a Project entity or use Project.executions repository')
        assert isinstance(self._project, entities.Project)
        return self._project

    @project.setter
    def project(self, project: entities.Project):
        if not isinstance(project, entities.Project):
            raise ValueError('Must input a valid Project entity')
        self._project = project

    def _list(self, filters: entities.Filters):
        """
        List resource executions

        :param dtlpy.entities.filters.Filters filters: dl.Filters entity to filters items
        :return:
        """
        url = "/executions/resource/query"

        # request
        success, response = self._client_api.gen_request(req_type='POST',
                                                         path=url,
                                                         json_req=filters.prepare())
        if not success:
            raise exceptions.PlatformException(response)

        return response.json()

    def list(self, filters: entities.Filters = None) -> entities.PagedEntities:
        """
        List resource executions

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param dtlpy.entities.filters.Filters filters: dl.Filters entity to filters items
        :return: Paged entity
        :rtype: dtlpy.entities.paged_entities.PagedEntities

        **Example**:

        .. code-block:: python

            item.resource_executions.list()
        """
        if filters is None:
            filters = entities.Filters(resource=entities.FiltersResource.RESOURCE_EXECUTION)
            if self._resource is not None:
                filters.add(field='resourceType', values=self._resource.__class__.__name__)
                filters.add(field='resourceId', values=self._resource.id)
            else:
                raise exceptions.PlatformException(
                    error='400',
                    message='Must have resource')
        # assert type filters
        elif not isinstance(filters, entities.Filters):
            raise exceptions.PlatformException(
                error='400',
                message='Unknown filters type: {!r}'.format(type(filters)))
        if filters.resource != entities.FiltersResource.RESOURCE_EXECUTION:
            raise exceptions.PlatformException(
                error='400',
                message='Filters resource must to be FiltersResource.RESOURCE_EXECUTION. '
                        'Got: {!r}'.format(filters.resource))
        if self._project is not None:
            filters.add(field='projectId', values=self._project.id)

        paged = entities.PagedEntities(items_repository=self,
                                       filters=filters,
                                       page_offset=filters.page,
                                       page_size=filters.page_size,
                                       client_api=self._client_api)
        paged.get_page()
        return paged

    def _build_entities_from_response(self, response_items) -> miscellaneous.List[entities.Execution]:
        pool = self._client_api.thread_pools(pool_name='entity.create')
        jobs = [None for _ in range(len(response_items))]
        # return execution list
        for i_item, item in enumerate(response_items):
            jobs[i_item] = pool.submit(entities.ResourceExecution._protected_from_json,
                                       **{'client_api': self._client_api,
                                          '_json': item,
                                          'project': self._project,
                                          'resource': self._resource})

        # get results
        results = [j.result() for j in jobs]
        # log errors
        _ = [logger.warning(r[1]) for r in results if r[0] is False]
        # return good jobs
        return miscellaneous.List([r[1] for r in results if r[0] is True])



import logging

from .. import exceptions, miscellaneous, entities, repositories, _api_reference
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')


class Assignments:
    """
    Assignments Repository

    The Assignments class allows users to manage assignments and their properties.
    Read more about `Task Assignment <https://developers.dataloop.ai/tutorials/task_workflows/create_a_task/chapter/>`_ in our Developers documentation.
    """

    def __init__(self,
                 client_api: ApiClient,
                 project: entities.Project = None,
                 task: entities.Task = None,
                 dataset: entities.Dataset = None,
                 project_id=None):
        self._client_api = client_api
        self._project = project
        self._dataset = dataset
        self._task = task

        self._project_id = project_id
        if self._project_id is None and self._project is not None:
            self._project_id = self._project.id

    ############
    # entities #
    ############
    @property
    def task(self) -> entities.Task:
        if self._task is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Missing "task". need to set an Task entity or use task.assignments repository')
        assert isinstance(self._task, entities.Task)
        return self._task

    @task.setter
    def task(self, task: entities.Task):
        if not isinstance(task, entities.Task):
            raise ValueError('Must input a valid Task entity')
        self._task = task

    @property
    def project_id(self):
        if self._project_id is not None:
            return self._project_id
        elif self._project is not None:
            return self._project.id
        else:
            return None

    @property
    def project(self) -> entities.Project:
        if self._project is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Missing "project". need to set a Project entity or use project.assignments repository')
        assert isinstance(self._project, entities.Project)
        return self._project

    @project.setter
    def project(self, project: entities.Project):
        if not isinstance(project, entities.Project):
            raise ValueError('Must input a valid Project entity')
        self._project = project

    ###########
    # methods #
    ###########
    def list(self,
             project_ids: list = None,
             status: str = None,
             assignment_name: str = None,
             assignee_id: str = None,
             pages_size: int = None,
             page_offset: int = None,
             task_id: int = None
             ) -> miscellaneous.List[entities.Assignment]:
        """
        Get Assignment list to be able to use it in your code.

        **Prerequisites**: You must be in the role of an *owner*, *developer*, or *annotation manager* who has been assigned as *owner* of the annotation task.

        :param list project_ids: search assignment by given list of project ids
        :param str status: search assignment by a given task status
        :param str assignment_name: search assignment by a given assignment name
        :param str assignee_id: the user email that assignee the assignment to it
        :param int pages_size: pages size of the output generator
        :param int page_offset: page offset of the output generator
        :param str task_id: search assignment by given task id
        :return: List of Assignment objects
        :rtype: miscellaneous.List[dtlpy.entities.assignment.Assignment]

        **Example**:

        .. code-block:: python

            assignments = task.assignments.list(status='complete', assignee_id='user@dataloop.ai', pages_size=100, page_offset=0)
        """

        # url
        url = '/assignments'

        query = list()
        if project_ids is not None:
            if not isinstance(project_ids, list):
                project_ids = [project_ids]
        elif self._project_id is not None:
            project_ids = [self._project_id]
        elif self._project is not None:
            project_ids = [self._project.id]
        else:
            raise exceptions.PlatformException(error='400', message='Must provide project')

        project_ids = ','.join(project_ids)
        query.append('projects={}'.format(project_ids))

        if status is not None:
            query.append('status={}'.format(status))
        if assignment_name is not None:
            query.append('name={}'.format(assignment_name))
        if assignee_id is not None:
            query.append('annotator={}'.format(assignee_id))
        if pages_size is not None:
            query.append('pageSize={}'.format(pages_size))
        if pages_size is None:
            query.append('pageSize={}'.format(500))
        if page_offset is not None:
            query.append('pageOffset={}'.format(page_offset))

        if task_id is None and self._task is not None:
            task_id = self._task.id
        if task_id is not None:
            query.append('taskId={}'.format(task_id))

        if len(query) > 0:
            query_string = '&'.join(query)
            url = '{}?{}'.format(url, query_string)

        success, response = self._client_api.gen_request(req_type='get',
                                                         path=url)
        if success:
            assignments = miscellaneous.List(
                [entities.Assignment.from_json(client_api=self._client_api,
                                               _json=_json, project=self._project, dataset=self._dataset,
                                               task=self._task)
                 for _json in response.json()['items']])
        else:
            logger.error('Platform error getting assignments')
            raise exceptions.PlatformException(response)
        return assignments

    @_api_reference.add(path='/assignments/{id}', method='get')
    def get(self,
            assignment_name: str = None,
            assignment_id: str = None):
        """
        Get Assignment object to use it in your code.

        :param str assignment_name: optional - search by name
        :param str assignment_id: optional - search by id
        :return: Assignment object
        :rtype: dtlpy.entities.assignment.Assignment

        **Example**:

        .. code-block:: python

            assignment = task.assignments.get(assignment_id='assignment_id')
        """

        if assignment_id is not None:
            url = '/assignments/{}'.format(assignment_id)
            success, response = self._client_api.gen_request(req_type='get',
                                                             path=url)
            if not success:
                raise exceptions.PlatformException('404', 'Assignment not found')
            else:
                assignment = entities.Assignment.from_json(_json=response.json(),
                                                           client_api=self._client_api,
                                                           project=self._project,
                                                           dataset=self._dataset,
                                                           task=self._task)
                # verify input assignment name is same as the given id
                if assignment_name is not None and assignment.name != assignment_name:
                    logger.warning(
                        "Mismatch found in assignments.get: assignment_name is different then assignment.name: "
                        "{!r} != {!r}".format(
                            assignment_name,
                            assignment.name))
        elif assignment_name is not None:
            assignments = [assignment for assignment in self.list() if assignment.name == assignment_name]
            if len(assignments) == 0:
                raise exceptions.PlatformException('404', 'Assignment not found')
            elif len(assignments) > 1:
                raise exceptions.PlatformException('404',
                                                   'More than one assignment exist with the same name: {}'.format(
                                                       assignment_name))
            else:
                assignment = assignments[0]
        else:
            raise exceptions.PlatformException('400', 'Must provide either assignment name or assignment id')

        assert isinstance(assignment, entities.Assignment)
        return assignment

    @property
    def platform_url(self):
        if self.task.id is None or self.project_id is None:
            raise ValueError("must have project and task")

        return self._client_api._get_resource_url(
            "projects/{}/tasks/{}/assignments".format(self.project_id, self.task.id))

    def open_in_web(self,
                    assignment_name: str = None,
                    assignment_id: str = None,
                    assignment: str = None):
        """
        Open the assignment in the platform.

        **Prerequisites**: All users.

        :param str assignment_name: the name of the assignment
        :param str assignment_id: the Id of the assignment
        :param dtlpy.entities.assignment.Assignment assignment: assignment object

        **Example**:

        .. code-block:: python

            task.assignments.open_in_web(assignment_id='assignment_id')
        """
        if assignment_name is not None:
            assignment = self.get(assignment_name=assignment_name)
        if assignment is not None:
            assignment.open_in_web()
        elif assignment_id is not None:
            self._client_api._open_in_web(url=self.platform_url + '/' + str(assignment_id))
        else:
            self._client_api._open_in_web(url=self.platform_url)

    @_api_reference.add(path='/assignments/{id}/reassign', method='post')
    def reassign(self,
                 assignee_id: str,
                 assignment: entities.Assignment = None,
                 assignment_id: str = None,
                 task: entities.Task = None,
                 task_id: str = None,
                 wait: bool = True):
        """
        Reassign an assignment.

        **Prerequisites**: You must be in the role of an *owner*, *developer*, or *annotation manager* who has been assigned as *owner* of the annotation task.

        :param str assignee_id: the email of the user that want to assign the assignment
        :param dtlpy.entities.assignment.Assignment assignment: assignment object
        :param assignment_id: the Id of the assignment
        :param dtlpy.entities.task.Task task: task object
        :param str task_id: the Id of the task that include the assignment
        :param bool wait: wait until reassign assignment finish
        :return: Assignment object
        :rtype: dtlpy.entities.assignment.Assignment

        **Example**:

        .. code-block:: python

            assignment = task.assignments.reassign(assignee_ids='annotator1@dataloop.ai')
        """
        if assignment_id is None and assignment is None:
            raise exceptions.PlatformException('400', 'Must provide either assignment or assignment_id')
        elif assignment_id is None:
            assignment_id = assignment.id

        if task_id is None and task is None:
            raise exceptions.PlatformException('400', 'Must provide either task or task_id')
        elif task_id is None:
            task_id = task.id

        url = '/assignments/{}/reassign'.format(assignment_id)

        payload = {
            'taskId': task_id,
            'annotator': assignee_id,
            'asynced': wait
        }

        success, response = self._client_api.gen_request(req_type='post',
                                                         path=url,
                                                         json_req=payload)
        if success:
            command = entities.Command.from_json(_json=response.json(),
                                                 client_api=self._client_api)
            if not wait:
                return command
            command = command.wait(timeout=0)
            if 'toAssignment' not in command.spec:
                raise exceptions.PlatformException(error='400',
                                                   message="'toAssignment' key is missing in command response: {}"
                                                   .format(response))
            return self.get(assignment_id=command.spec['toAssignment'])
        else:
            raise exceptions.PlatformException(response)

    @_api_reference.add(path='/assignments/{id}/redistribute', method='post')
    def redistribute(self,
                     workload: entities.Workload,
                     assignment: entities.Assignment = None,
                     assignment_id: str = None,
                     task: entities.Task = None,
                     task_id: str = None,
                     wait: bool = True):
        """
        Redistribute an assignment.

        **Prerequisites**: You must be in the role of an *owner*, *developer*, or *annotation manager* who has been assigned as *owner* of the annotation task.

        **Example**:

        :param dtlpy.entities.assignment.Workload workload: list of WorkloadUnit objects. Customize distribution (percentage) between the task assignees. For example: [dl.WorkloadUnit(annotator@hi.com, 80), dl.WorkloadUnit(annotator2@hi.com, 20)]
        :param dtlpy.entities.assignment.Assignment assignment: assignment object
        :param str assignment_id: the Id of the assignment
        :param dtlpy.entities.task.Task task: the task object that include the assignment
        :param str task_id: the Id of the task that include the assignment
        :param bool wait: wait until redistribute assignment finish
        :return: Assignment object
        :rtype: dtlpy.entities.assignment.Assignment assignment

        .. code-block:: python

            assignment = task.assignments.redistribute(workload=dl.Workload([dl.WorkloadUnit(assignee_id="annotator1@dataloop.ai", load=50),
                                                                dl.WorkloadUnit(assignee_id="annotator2@dataloop.ai", load=50)]))
        """
        if assignment_id is None and assignment is None:
            raise exceptions.PlatformException('400', 'Must provide either assignment or assignment_id')
        elif assignment_id is None:
            assignment_id = assignment.id

        if task_id is None and task is None:
            raise exceptions.PlatformException('400', 'Must provide either task or task_id')
        elif task_id is None:
            task_id = task.id

        url = '/assignments/{}/redistribute'.format(assignment_id)

        payload = {
            'taskId': task_id,
            'workload': workload.to_json(),
            'asynced': wait
        }

        if self._task is None:
            self._task = self.get(assignment_id=assignment_id).task
        if task is None:
            task = self._task

        success, response = self._client_api.gen_request(req_type='post',
                                                         path=url,
                                                         json_req=payload)
        if success:
            command = entities.Command.from_json(_json=response.json(),
                                                 client_api=self._client_api)
            if not wait:
                return command
            command = command.wait(timeout=0)
            if 'workload' not in command.spec:
                raise exceptions.PlatformException(error='400',
                                                   message="workload key is missing in command response: {}"
                                                   .format(response))

            task_assignments = task.assignments.list()
            workers = list()
            for worker in workload:
                workers.append(worker.assignee_id.lower())

            redistributed_assignments = list()
            for ass in task_assignments:
                if ass.annotator in workers:
                    redistributed_assignments.append(ass)
                    workers.remove(ass.annotator)
                    if not workers:
                        break

            return miscellaneous.List(redistributed_assignments)
        else:
            raise exceptions.PlatformException(response)

    @_api_reference.add(path='/assignments/{id}', method='patch')
    def update(self,
               assignment: entities.Assignment = None,
               system_metadata: bool = False
               ) -> entities.Assignment:
        """
        Update an assignment.

        **Prerequisites**: You must be in the role of an *owner*, *developer*, or *annotation manager* who has been assigned as *owner* of the annotation task.

        :param dtlpy.entities.assignment.Assignment assignment assignment: assignment entity
        :param bool system_metadata: True, if you want to change metadata system
        :return: Assignment object
        :rtype: dtlpy.entities.assignment.Assignment assignment

        **Example**:

        .. code-block:: python

            assignment = task.assignments.update(assignment='assignment_entity', system_metadata=False)
        """
        url = '/assignments/{}'.format(assignment.id)

        if system_metadata:
            url += '?system=true'

        success, response = self._client_api.gen_request(req_type='patch',
                                                         path=url,
                                                         json_req=assignment.to_json())
        if success:
            return entities.Assignment.from_json(_json=response.json(),
                                                 client_api=self._client_api, project=self._project,
                                                 dataset=self._dataset, task=self._task)
        else:
            raise exceptions.PlatformException(response)

    def create(self,
               assignee_id: str,
               task: entities.Task = None,
               filters: entities.Filters = None,
               items: list = None) -> entities.Assignment:
        """
        Create a new assignment.

        **Prerequisites**: You must be in the role of an *owner*, *developer*, or *annotation manager* who has been assigned as *owner* of the annotation task.

        :param str assignee_id: the email of the user that want to assign the assignment
        :param dtlpy.entities.task.Task task: the task object that include the assignment
        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :param list items: list of items (item Id or objects) to insert to the assignment
        :return: Assignment object
        :rtype: dtlpy.entities.assignment.Assignment assignment

        **Example**:

        .. code-block:: python

            assignment = task.assignments.create(assignee_id='annotator1@dataloop.ai')
        """
        return self._create_in_task(assignee_id=assignee_id, task=task, filters=filters, items=items)

    def _create_in_task(self, assignee_id, task, filters=None, items=None) -> entities.Assignment:

        if task is None:
            if self._task is None:
                raise exceptions.PlatformException('400', 'Must provide task')
            task = self._task

        assignments_before = [ass.id for ass in task.assignments.list()]

        if filters is None and items is None:
            raise exceptions.PlatformException('400', 'Must provide either filters or items list')

        workload = entities.Workload.generate(assignee_ids=[assignee_id])
        task = task.add_items(filters=filters, items=items, workload=workload, limit=None)
        assignments = [ass for ass in task.assignments.list() if ass.id not in assignments_before]

        if len(assignments) < 1:
            raise exceptions.PlatformException('Error creating an assignment, '
                                               'Please use task.add_items() to perform this action')

        return assignments[0]

    def __item_operations(self, dataset: entities.Dataset,
                          op, assignment_id=None, assignment_name=None, filters=None, items=None):
        if assignment_id is None and assignment_name is None:
            raise exceptions.PlatformException('400', 'Must provide either assignment name or assignment id')
        elif assignment_id is None:
            assignment_id = self.get(assignment_name=assignment_name).id

        try:
            if filters is None and items is None:
                raise exceptions.PlatformException('400', 'Must provide either filters or items list')

            if filters is None:
                if not isinstance(items, list):
                    items = [items]
                filters = entities.Filters(field='id',
                                           values=[item.id for item in items],
                                           operator=entities.FiltersOperations.IN,
                                           use_defaults=False)

            filters._ref_assignment = True
            filters._ref_assignment_id = assignment_id
            filters._ref_op = op

            return dataset.items.update(filters=filters)
        finally:
            if filters is not None:
                filters._nullify_refs()

    def get_items(self,
                  assignment: entities.Assignment = None,
                  assignment_id=None,
                  assignment_name=None,
                  dataset=None,
                  filters=None
                  ) -> entities.PagedEntities:
        """
        Get all the items in the assignment.

        **Prerequisites**: You must be in the role of an *owner*, *developer*, or *annotation manager* who has been assigned as *owner* of the annotation task.

        :param dtlpy.entities.assignment.Assignment assignment: assignment object
        :param assignment_id: the Id of the assignment
        :param str assignment_name: the name of the assignment
        :param dtlpy.entities.dataset.Dataset dataset: dataset object, the dataset that refer to the assignment
        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :return: pages of the items
        :rtype: dtlpy.entities.paged_entities.PagedEntities

        **Example**:

        .. code-block:: python

            items = task.assignments.get_items(assignment_id='assignment_id')
        """
        if assignment is None and assignment_id is None and assignment_name is None:
            raise exceptions.PlatformException('400',
                                               'Please provide either assignment,  assignment_id or assignment_name')

        if assignment_id is None:
            if assignment is None:
                assignment = self.get(assignment_name=assignment_name)
            assignment_id = assignment.id

        if dataset is None and self._dataset is None:
            if assignment is None:
                assignment = self.get(assignment_id=assignment_id, assignment_name=assignment_name)
            if assignment.dataset_id is None:
                raise exceptions.PlatformException('400', 'Please provide a dataset entity')
            dataset = repositories.Datasets(client_api=self._client_api, project=self._project).get(
                dataset_id=assignment.dataset_id)
        elif dataset is None:
            dataset = self._dataset

        if filters is None:
            filters = entities.Filters(use_defaults=False)
        filters.add(field='metadata.system.refs.id', values=[assignment_id], operator=entities.FiltersOperations.IN)

        return dataset.items.list(filters=filters)

    def set_status(self,
                   status: str,
                   operation: str,
                   item_id: str,
                   assignment_id: str
                   ) -> bool:
        """
        Set item status within assignment.

        **Prerequisites**: You must be in the role of an *owner*, *developer*, or *annotation manager* who has been assigned as *owner* of the annotation task.

        :param str status: string the describes the status
        :param str operation: the status action need 'create' or 'delete'
        :param str item_id: item id that want to set his status
        :param assignment_id: the Id of the assignment
        :return: True id success
        :rtype: bool

        **Example**:

        .. code-block:: python

            success = task.assignments.set_status(assignment_id='assignment_id',
                                        status='complete',
                                        operation='created',
                                        item_id='item_id')
        """
        url = '/assignments/{assignment_id}/items/{item_id}/status'.format(assignment_id=assignment_id, item_id=item_id)
        payload = {
            'operation': operation,
            'status': status
        }
        success, response = self._client_api.gen_request(
            req_type='post',
            path=url,
            json_req=payload
        )

        if not success:
            raise exceptions.PlatformException(response)

        return True
